### Wrapper API

from metient.lib import migration_history_inference as mig_hist
from metient.util import plotting_util as plutil

def evaluate(tree_fn, tsv_fn, weights, print_config, output_dir, run_name, 
             O=None, sample_size=-1, solve_polytomies=False, num_runs=1):
    '''
    Runs Metient-evaluate, and infers the observed clone percentages and the labels of the clone tree.

    Args: 
        REQUIRED:
        tree_fn: Path to .txt file where each line is an edge from the first index to the second index. Must correspond to the cluster_index column in the input tsv.
        tsv_fn: Path to .tsv file where each row in this tsv should correspond to a single mutation/mutation cluster in a single tumor sample. Required columns:
                anatomical_site_index, anatomical_site_label, cluster_index, character_index, character_label, ref, var, var_read_prob, site_category
        weights: Weights object which specifies the relative weighting to place on each part of the objective
        print_config: PrintConfig object which specifies saving/visualization configuration
        output_dir: Path for where to save outputs
        run_name: Name for this patient which will be used to save all outputs.

        OPTIONAL:
        O: a dictionary mapping anatomical site name (as used in tsv_fn) -> frequency of metastasis (these values should be normalized)
        sample_size: how many samples to have Metient solve in parallel
        solve_polytomies: bool, whether or not to resolve polytomies 

    Outputs migration history inferences for a single patient.
    '''
    return mig_hist.evaluate(tree_fn, tsv_fn, weights, print_config, output_dir, run_name, 
                           O=O, sample_size=sample_size, solve_polytomies=solve_polytomies, 
                           num_runs=num_runs)

def evaluate_label_clone_tree(tree_fn, tsv_fn, weights, print_config, output_dir, run_name, 
                            O=None, sample_size=-1, solve_polytomies=False, num_runs=1, bias_weights=True):
    '''
    Runs Metient-evaluate with observed clone percentages inputted, and only inferring the labels of the clone tree.

    Args: 
        REQUIRED:
        tree_fn: Path to .txt file where each line is an edge from the first index to the second index. Must correspond to the cluster_index column in the input tsv.
        tsv_fn: Path to .tsv file where each row in this tsv should correspond to a single mutation/mutation cluster in a single tumor sample. Required columns:
                anatomical_site_index, anatomical_site_label, cluster_index, cluster_label, present, site_category, num_mutations
        weights: Weights object which specifies the relative weighting to place on each part of the objective
        print_config: PrintConfig object which specifies saving/visualization configuration
        output_dir: Path for where to save outputs
        run_name: Name for this patient which will be used to save all outputs.

        OPTIONAL:
        O: a dictionary mapping anatomical site name (as used in tsv_fn) -> frequency of metastasis (these values should be normalized)
        sample_size: how many samples to have Metient solve in parallel
        solve_polytomies: bool, whether or not to resolve polytomies 
    
    Outputs migration history inferences for a single patient.
    '''
    return mig_hist.evaluate_label_clone_tree(tree_fn, tsv_fn, weights, print_config, output_dir, run_name, 
                                            O=O, sample_size=sample_size, bias_weights=bias_weights, 
                                            solve_polytomies=solve_polytomies, num_runs=num_runs)

def calibrate(tree_fns, tsv_fns, print_config, output_dir, run_names, 
              Os=None, sample_size=-1, solve_polytomies=False, num_runs=1):
    '''
    Runs Metient-calibrate on a cohort of patients. For each patient, we infer the observed clone percentages and the labels of the clone tree.

    Args: 
        REQUIRED:
        tree_fns: List of paths to .txt files. In each .txt file, each line is an edge from the first index to the second index. Must correspond to the cluster_index column in the input tsv.
        tsv_fns: List of paths to .tsv files. In each .tsv file, each row in this tsv should correspond to a single mutation/mutation cluster in a single tumor sample. Required columns:
                anatomical_site_index, anatomical_site_label, cluster_index, character_index, character_label, ref, var, var_read_prob, site_category
        print_config: PrintConfig object which specifies saving/visualization configuration
        output_dir: Path for where to save outputs
        run_names: List of patient names which will be used to save all outputs.

        NOTE: tree_fns[i] and tsv_fns[i] and run_names[i] all correspond to patient i.

        OPTIONAL:
        O: a dictionary mapping anatomical site name (as used in tsv_fn) -> frequency of metastasis (these values should be normalized)
        sample_size: how many samples to have Metient solve in parallel
        solve_polytomies: bool, whether or not to resolve polytomies 

    Outputs migration history inferences for a full cohort.
    '''
    return mig_hist.calibrate(tree_fns, tsv_fns, print_config, output_dir, run_names, 
                            Os=Os, sample_size=sample_size, bias_weights=True, 
                            solve_polytomies=solve_polytomies, num_runs=num_runs)

def calibrate_label_clone_tree(tree_fns, tsv_fns, print_config, output_dir, run_names, 
                             Os=None, sample_size=-1, solve_polytomies=False, num_runs=1):
    '''
    Runs Metient-calibrate on a cohort of patients. For each patient, we use the inputted observed clone percentages, and only infer the labels of the clone tree.

    Args: 
        REQUIRED:
        tree_fns: List of paths to .txt files. In each .txt file, each line is an edge from the first index to the second index. Must correspond to the cluster_index column in the input tsv.
        tsv_fns: List of paths to .tsv files. In each .tsv file, each row in this tsv should correspond to a single mutation/mutation cluster in a single tumor sample. Required columns:
                anatomical_site_index, anatomical_site_label, cluster_index, cluster_label, present, site_category, num_mutations
        print_config: PrintConfig object which specifies saving/visualization configuration
        output_dir: Path for where to save outputs
        run_names: List of patient names which will be used to save all outputs.

        NOTE: tree_fns[i] and tsv_fns[i] and run_names[i] all correspond to patient i.

        OPTIONAL:
        O: a dictionary mapping anatomical site name (as used in tsv_fn) -> frequency of metastasis (these values should be normalized)
        sample_size: how many samples to have Metient solve in parallel
        solve_polytomies: bool, whether or not to resolve polytomies 

    Outputs migration history inferences for a full cohort.
    '''
    return mig_hist.calibrate_label_clone_tree(tree_fns, tsv_fns, print_config, output_dir, run_names, 
                                             Os=Os, sample_size=sample_size, bias_weights=True, 
                                             solve_polytomies=solve_polytomies, num_runs=num_runs)


class PrintConfig:
    def __init__(self, visualize=True, verbose=False, k_best_trees=float("inf"), save_outputs=True, custom_colors=None):
        '''
        Args:
            visualize: bool, whether to visualize loss, best tree, and migration graph
            verbose: bool, whether to print debug info
            k_best_trees: int, number of best tree solutions to visualize (if 1, only show best tree)
            save_outputs: bool, whether to save pngs and pickle files
            custom_colors: array of hex strings (with length = number of anatomical sites) to be used as custom colors in output visualizations
        '''
        if k_best_trees <= 0:
            raise ValueError("k_best_trees must be >= 1")
        self.visualize = visualize
        self.verbose = verbose 
        self.k_best_trees = k_best_trees
        self.save_outputs = save_outputs
        self.custom_colors = custom_colors

class Weights:
    def __init__(self, mig=0.48, comig=0.30, seed_site=0.22, gen_dist=0.0, organotrop=0.0, data_fit=15.0, reg=0.5, entropy=0.01):
        '''
        The higher the inputted weight, the higher the penalty on that metric.

        Args:
            mig: weight to place on migration number. Default is based on calibration to real data.
            comig: weight to place on comigration number. Default is based on calibration to real data.
            seed_site: weight to place on seeding site number. Default is based on calibration to real data.
            gen_dist: weight to place on genetic distance loss.
            organotrop: weight to place on organotropism loss.
            data_fit: weight to place on negative log likelihood loss of observed clone percentages.
            reg: weight to place on regularization loss for observed clone percentages.
            entropy: weight to place on negative entropy.
        '''
        self.data_fit = data_fit
        self.mig = mig
        self.comig = comig
        self.seed_site = seed_site
        self.reg = reg
        self.gen_dist = gen_dist
        self.organotrop = organotrop
        self.entropy = entropy

def migration_graph(V, A):
    '''
    V: Vertex labeling matrix where columns are one-hot vectors representing the
    anatomical site that the node originated from (num_sites x num_nodes)
    A:  Adjacency matrix (directed) of the full tree (num_nodes x num_nodes)
    '''
    return plutil.migration_graph(V, A)

def seeding_pattern(V, A):
    '''
    V: Vertex labeling matrix where columns are one-hot vectors representing the
    anatomical site that the node originated from (num_sites x num_nodes)
    A:  Adjacency matrix (directed) of the full tree (num_nodes x num_nodes)

    Returns: verbal description of the seeding pattern, one of:
    {monoclonal, polyclonal} {single-source, multi-source, reseeding}
    '''
    return plutil.seeding_pattern(V, A)

def phyleticity(V, A, node_info):
    '''
    V: Vertex labeling matrix where columns are one-hot vectors representing the
    anatomical site that the node originated from (num_sites x num_nodes)
    A:  Adjacency matrix (directed) of the full tree (num_nodes x num_nodes)
    node_info: Dictionary mapping node indices to a tuple (label (str), is_witness_node (bool), is_polytomy_resolver_node (bool)). 
               This is outputted in Metient's pkl.gz file

    After determining which nodes perform seeding (i.e., nodes which have a different
    color from their parent), if all nodes can be reached from the highest level node 
    in the seeding clusters (closest to root), returns monophyletic, else polyphyletic
    '''
    return plutil.phyleticity(V, A, node_info)

def seeding_clusters(V, A, node_info):
    '''
    V: Vertex labeling matrix where columns are one-hot vectors representing the
    anatomical site that the node originated from (num_sites x num_nodes)
    A:  Adjacency matrix (directed) of the full tree (num_nodes x num_nodes)
    node_info: Dictionary mapping node indices to a tuple (label (str), is_witness_node (bool), is_polytomy_resolver_node (bool)). 
               This is outputted in Metient's pkl.gz file

    Returns: list of nodes whose parent is a different color than itself
    '''
    return plutil.seeding_clusters(V, A, node_info)

def site_clonality(V, A):
    '''
    V: Vertex labeling matrix where columns are one-hot vectors representing the
    anatomical site that the node originated from (num_sites x num_nodes)
    A:  Adjacency matrix (directed) of the full tree (num_nodes x num_nodes)

    Returns monoclonal if every site is seeded by one clone,
    else returns polyclonal.
    '''
    return plutil.site_clonality(V, A)

def genetic_clonality(V, A, node_info):
    '''
    V: Vertex labeling matrix where columns are one-hot vectors representing the
    anatomical site that the node originated from (num_sites x num_nodes)
    A:  Adjacency matrix (directed) of the full tree (num_nodes x num_nodes)
    node_info: Dictionary mapping node indices to a tuple (label (str), is_witness_node (bool), is_polytomy_resolver_node (bool)). 
               This is outputted in Metient's pkl.gz file

    Returns monoclonal if every site is seeded by the *same* clone,
    else returns polyclonal.
    '''
    return plutil.genetic_clonality(V, A, node_info)


import pickle
import glob
import heapq
import torch
import os
import re
import pandas as pd
import joblib
import gzip
import seaborn as sns
import matplotlib.pyplot as plt
import torch
import torch.optim as optim
import numpy as np

from metient.util.globals import *
from metient.util import plotting_util as plot_util
from metient.util import vertex_labeling_util as vutil

def plot_cross_ent_chart(data_dict, output_dir):
    fig = plt.figure(figsize=(2.5, 3),dpi=200)
    
    keys = list(data_dict.keys())
    sorted_keys = sorted(keys, key=lambda x: x[0], reverse=True)
    string_keys = [str((round(k[0], 2),round(k[1], 2),round(k[2], 2))) for k in sorted_keys]
    values = [-1*data_dict[k].item() for k in sorted_keys]
    snsfig = sns.barplot(x=string_keys, y=values, palette=sns.color_palette("muted"),
                         order=string_keys)

    snsfig.spines['top'].set_visible(False)
    snsfig.spines['right'].set_visible(False)

    plt.xlabel("Theta: (migration weight, seeding site weight, \nmigration delta weight)",fontsize=10)
    plt.ylabel("-E(theta)",fontsize=15)
    plt.xticks(fontsize=8, rotation=45)
    plt.yticks(fontsize=12)
    plt.ylim(min(values)-5, max(values)+5)
    fig.savefig(os.path.join(output_dir, f"cross_entropy_distribution.png"), dpi=600, bbox_inches='tight', pad_inches=0.5) 
    plt.close()

def stable_softmax(x):
    # Subtract the maximum value for numerical stability
    x_max, _ = torch.max(x, dim=0, keepdim=True)
    exp_x = torch.exp(x - x_max)
    return exp_x / torch.sum(exp_x, dim=0)

def cross_ent(loss_dicts, pt_weight, thetas, tau):
    '''
    Computes the cross entropy between the target distribution (genetic distance)
    and the predicted distribution (parsimony) using the input thetas (which we are
    trying to optimize)
    '''
    theta_X = torch.zeros((len(loss_dicts)))
    gen_dist_scores = torch.zeros((len(loss_dicts)))
    organotrop_scores = torch.zeros((len(loss_dicts)))
    for i, loss_dict in enumerate(loss_dicts):
        m = loss_dict[MIG_KEY]
        c = loss_dict[COMIG_KEY]
        s = loss_dict[SEEDING_KEY]
        g = loss_dict[GEN_DIST_KEY]
        o = loss_dict[ORGANOTROP_KEY]
        theta_X[i] = -1.0*(thetas[0]*m + thetas[1]*c + thetas[2]*s)
        gen_dist_scores[i] = -tau*g
        organotrop_scores[i] = -tau*o
    #print()
    # print("theta x", theta_X, "\n gen dist", gen_dist_scores, "\n organotrop", organotrop_scores)
    # this is in the case where there is no seeding detected (happens rarely in some datasets)
    if torch.sum(theta_X) == 0 or torch.isnan(gen_dist_scores).any():
        return 0.0
    theta_X = stable_softmax(theta_X)

    log_wn = np.log2(pt_weight+1) # So that weight=1 doesn't go to 0
    
    cross_ent_sum = 0.0
    if not torch.sum(gen_dist_scores == 0):
        gen_dist_scores = stable_softmax(gen_dist_scores)
        cross_ent_sum += -log_wn*torch.sum(torch.mul(gen_dist_scores, torch.log2(theta_X+0.1)))
    if not torch.sum(organotrop_scores == 0):
        organotrop_scores = stable_softmax(organotrop_scores)
        cross_ent_sum += -log_wn*torch.sum(torch.mul(organotrop_scores, torch.log2(theta_X+0.1)))

    return cross_ent_sum

def get_pickle_filenames(pickle_files_dirs, suffix=None):
    pickle_filenames = []
    match = f"{suffix}.pkl.gz" if suffix != None else ".pkl.gz" 
    for pickle_files_dir in pickle_files_dirs:
        for file in os.listdir(pickle_files_dir):
            if match in file:
                pickle_filenames.append(os.path.join(pickle_files_dir, file))
    return pickle_filenames
    
def get_max_cross_ent_thetas(pickle_file_list, patient_weights, tau=3.0, use_min_tau=False):
    '''
    pickle_file_list: list of paths to Metient results
    patient_weights: the 

    Returns the parsimony weights which give the best cross entropy 
    between the target distribution (genetic distance) and predicted 
    distribution (parsimony) across all patients in pickle_files_dir
    '''
    
    min_tau = float("inf")
    all_data = []
    for pkl_file,pt_weight in zip(pickle_file_list, patient_weights):
        with gzip.open(pkl_file,'rb') as f:
            pckl = pickle.load(f)
        
        loss_dicts = pckl[OUT_LOSS_DICT_KEY]
        pars_metrics = set()
        for loss_dict in loss_dicts:
            pars_metrics.add((loss_dict[MIG_KEY],loss_dict[COMIG_KEY],loss_dict[SEEDING_KEY]))
        
        # This patient won't change the cross entropy loss
        if len(pars_metrics) == 1:
            continue

        all_data.append((loss_dicts, pt_weight))

        if use_min_tau:
            gen_dist_scores = torch.zeros((len(loss_dicts)))
            for i, loss_dict in enumerate(loss_dicts):
                gen_dist_scores[i] = loss_dict[GEN_DIST_KEY]
            unique_gs,_ = torch.sort(torch.unique(gen_dist_scores))
            if len(unique_gs) >= 2:
                print("gen dist diff", (unique_gs[1]-unique_gs[0]))
                min_tau = min(min_tau, (unique_gs[1]-unique_gs[0]))
    
    if len(all_data) == 0:
        print("WARNING: Unable to calibrate since no patients have multiple Pareto optimal trees with multiple unique Pareto metrics. Using default weighting, wm > wc > ws")
        return [0.46, 0.29, 0.25] # these are estimated from multiple cancer cohorts
    
    if use_min_tau:
        tau = min_tau
        print("min_tau", tau)
        
    print(f"Calibrating to {len(all_data)} patients")

    patience = 50
    min_delta = 0.00005
    current_patience = 0
    best_loss = float('inf')
    
    thetas = torch.tensor([1.0,1.0,1.0], requires_grad=True)
    optimizer = optim.SGD([thetas], lr=0.01)

    max_iter = 1000
    losses = []
    for step in range(max_iter):
        optimizer.zero_grad()
        total_cross_ent = 0.0
        for i, (patient_loss_dicts, pt_weight) in enumerate(all_data):
            patient_cross_ent = cross_ent(patient_loss_dicts, pt_weight, thetas, tau)
            total_cross_ent += patient_cross_ent
        total_cross_ent = total_cross_ent/float(len(all_data))
        total_cross_ent.backward()
        optimizer.step()
        losses.append(float(total_cross_ent))

        if step % 100 == 0:
            pass

        # check if loss improved
        if total_cross_ent < best_loss - min_delta:
            best_loss = total_cross_ent
            current_patience = 0
        else:
            current_patience += 1

        # check if early stopping criteria met
        if current_patience >= patience:
            print(f"Early stopping after {step+1} epochs.")
            max_iter = step + 1
            break

    thetas = stable_softmax(thetas)

    print(f"Optimized thetas: {thetas}")

    # import matplotlib.pyplot as plt
    # import numpy as np
    # fig = plt.figure(figsize=(2, 2),dpi=200)
    # plt.plot(range(1, max_iter + 1), losses, label='Training Loss')
    # plt.title('Training Loss Over Iterations')
    # plt.xlabel('Iterations')
    # plt.ylabel('Loss')
    # plt.legend()
    # plt.show()

    return [float(thetas[0]), float(thetas[1]), float(thetas[2])]



######### Methods taken directly from MACHINA repo #########

def load_machina_results(machina_results_dir):
    # Taken directly from https://github.com/raphael-group/machina/blob/master/result/sims/simulations_1.ipynb
    files_new_m5=['results_MACHINA_m5.txt',]
    files_new_m8=['results_MACHINA_m8.txt', ]
    res_m5 = pd.concat([pd.read_csv(os.path.join(machina_results_dir, filename)) for filename in files_new_m5]).reindex()
    res_m8 = pd.concat([pd.read_csv(os.path.join(machina_results_dir, filename)) for filename in files_new_m8]).reindex()
    res_m5 = res_m5[(res_m5['enforced']=='R') | (res_m5['enforced'].isnull())]
    res_m8 = res_m8[(res_m8['enforced']=='R') | (res_m8['enforced'].isnull())]
    res_m5 = res_m5.replace({'pattern': {'S': 'pS', 'M' : 'pM', 'R' : 'pR'}})
    res_m8 = res_m8.replace({'pattern': {'S': 'pS', 'M' : 'pM', 'R' : 'pR'}})
    res_m8_MACHINA = res_m8[res_m8['method'] == 'MACHINA'].replace({'inferred': {'pPS': 'pS', 'mPS' : 'mS'}})
    res_m5_MACHINA = res_m5[res_m5['method'] == 'MACHINA'].replace({'inferred': {'pPS': 'pS', 'mPS' : 'mS'}})

    return res_m5_MACHINA, res_m8_MACHINA

def load_machina_results_new_split(machina_results_dir):
    gt_df = pd.read_csv("/data/morrisq/divyak/projects/metient/metient/data/machina_sims/gt_pattern.csv")
    gt_df['site'] = gt_df['site'].astype(str)
    gt_df['mig_type'] = gt_df['mig_type'].astype(str)
    gt_df['seed'] = gt_df['seed'].astype(str)
    files_new_m5=['results_MACHINA_m5.txt',]
    files_new_m8=['results_MACHINA_m8.txt', ]
    res_m5 = pd.concat([pd.read_csv(os.path.join(machina_results_dir, filename)) for filename in files_new_m5]).reindex()
    res_m8 = pd.concat([pd.read_csv(os.path.join(machina_results_dir, filename)) for filename in files_new_m8]).reindex()
    res_m5 = res_m5[(res_m5['enforced']=='R') | (res_m5['enforced'].isnull())]
    res_m8 = res_m8[(res_m8['enforced']=='R') | (res_m8['enforced'].isnull())]
    
    res_m8_MACHINA = res_m8[res_m8['method'] == 'MACHINA']
    res_m5_MACHINA = res_m5[res_m5['method'] == 'MACHINA']
    
    res_m8_MACHINA = res_m8_MACHINA.replace({'pattern': {'pS': 'S', 'pM' : 'M', 'pR' : 'R'}})
    res_m5_MACHINA = res_m5_MACHINA.replace({'pattern': {'pS': 'S', 'pM' : 'M', 'pR' : 'R'}})
    res_m5_MACHINA['new_gt_pattern'] = res_m5_MACHINA.apply(lambda row: gt_df[(gt_df['site']=='m5')&(gt_df['mig_type']==row['pattern'])&(gt_df['seed']==str(row['seed']))]['gt_pattern'].item(), axis=1)
    res_m8_MACHINA['new_gt_pattern'] = res_m8_MACHINA.apply(lambda row: gt_df[(gt_df['site']=='m8')&(gt_df['mig_type']==row['pattern'])&(gt_df['seed']==str(row['seed']))]['gt_pattern'].item(), axis=1)

    return res_m5_MACHINA, res_m8_MACHINA

# Taken from MACHINA 
def get_mutations(edge_list, u):
    # find parent
    s = re.split("_|;|^", u)
    mutations = set()
    for i in s:
        if i.isdigit():
            mutations.add(int(i))

    for edge in edge_list:
        uu = edge[0]
        vv = edge[1]
        if vv == u:
            return mutations | get_mutations(edge_list, uu)

    return mutations

# Taken from MACHINA 
def parse_clone_tree(filename_T, filename_l):
    edges = []
    with open(filename_T) as f:
        for line in f:
            s = line.rstrip("\n").split(" ")
            edges += [(s[0], s[1])]

    labeling = {}
    with open(filename_l) as f:
        for line in f:
            s = line.rstrip("\n").split(" ")
            labeling[s[0]] = s[1]

    # find migration edges
    migration_edges = []
    for (u, v) in edges:
        if labeling[u] != labeling[v]:
            migration_edges += [(u,v)]

    return edges, migration_edges,labeling

# Taken from MACHINA definition
def identify_seeding_clones_mach(edge_list, migration_edge_list):
    res = set()
    for (u,v) in migration_edge_list:
        muts_u = get_mutations(edge_list, u)
        muts_v = get_mutations(edge_list, v)
        res.add(frozenset(muts_u))
    return res

# Metient definition of seeding clones
def identify_seeding_clones_met(edge_list, migration_edge_list):
    res = set()
    for (u,v) in migration_edge_list:
        muts_v = get_mutations(edge_list, v)
        res.add(frozenset(muts_v))
    return res

# Taken from MACHINA 
def parse_migration_graph(filename_G):
    edges = []
    with open(filename_G) as f:
        for line in f:
            s = line.rstrip("\n").split(" ")
            edges += [(s[0], s[1])]
    
    return edges

# Taken from MACHINA 
def multi_graph_to_set(edge_list):
    count = {}
    res = set()
    for edge in edge_list:
        if edge not in count:
            count[edge] = 1
        else:
            count[edge] += 1
        res.add((edge[0], edge[1], count[edge]))
    return res

########### Methods for evaluating Metient ouputs #############

def create_sparse_tensor(sparse_info):
    """
    Creates a sparse tensor from a tuple containing indices, values, and size.

    Args:
        sparse_info (tuple): A tuple of the form (indices, values, size) where:
            - indices: A 2D tensor of shape (2, N) representing the indices of the non-zero values.
            - values: A 1D tensor of shape (N,) representing the non-zero values.
            - size: A torch.Size object representing the size of the sparse tensor.

    Returns:
        torch.sparse.FloatTensor: A sparse tensor created from the input information.
    """
    indices, values, size = sparse_info
    return torch.sparse.FloatTensor(indices, values, size)

def metient_parse_clone_tree(results_dict, met_tree_num):
    V = torch.tensor(results_dict[OUT_LABElING_KEY][met_tree_num])
    A = create_sparse_tensor(results_dict[OUT_ADJ_KEY][met_tree_num])
    idx_to_lbl = results_dict[OUT_IDX_LABEL_KEY][met_tree_num]
    idx_to_string_label = {i:";".join(idx_to_lbl[i][0]) for i in idx_to_lbl}

    edges = [("GL", '0')]
    for i, j in vutil.tree_iterator(A):
        edges.append((idx_to_string_label[i], idx_to_string_label[j]))
    
    migration_edges = []
    for i, j in vutil.tree_iterator(A):
        if torch.argmax(V[:,i]) != torch.argmax(V[:,j]):
            migration_edges.append((idx_to_string_label[i], idx_to_string_label[j]))
    return edges, migration_edges

def metient_parse_mig_graph(results_dict, met_tree_num):
    V = torch.tensor(results_dict[OUT_LABElING_KEY][met_tree_num])
    A = create_sparse_tensor(results_dict[OUT_ADJ_KEY][met_tree_num])
    sites = results_dict[OUT_SITES_KEY]
    G = plot_util.migration_graph(V, A)
    migration_edges = []
    for i, row in enumerate(G):
        for j, val in enumerate(row):
            if val != 0:
                for _ in range(int(val)):
                    migration_edges.append((sites[i], sites[j]))
    return migration_edges

class HeapTree:
    def __init__(self, loss, results_dict, met_tree_num, tree_num):
        self.loss = loss
        self.results_dict = results_dict
        self.met_tree_num = met_tree_num
        self.tree_num = tree_num

   # override the comparison operator
    def __lt__(self, other):
        return self.loss < other.loss
    
def get_metient_min_loss_trees(site_mig_type_dir, seed, k, loss_thres=1.0, suffix=""):
    # Get all clone trees for the seed
    tree_pickles = glob.glob(os.path.join(site_mig_type_dir, f"*_seed{seed}{suffix}.pkl.gz"))
    
    # Keep track of the best trees and losses
    min_heap = [] 
    for tree_pickle in tree_pickles:
        clone_tree_num = os.path.basename(tree_pickle).replace("tree", "").replace(f"_seed{seed}{suffix}.pkl.gz", "")
        with gzip.open(tree_pickle,'rb') as f:
            results_dict = joblib.load(f)
        for met_tree_num, loss in enumerate(results_dict['losses']):
            heapq.heappush(min_heap, HeapTree(loss, results_dict, met_tree_num, clone_tree_num))
    out = []
    min_loss = min_heap[0].loss
    while len(out) < k and len(min_heap) > 0:
        item = heapq.heappop(min_heap)
        # only add tree if it's ~= to the min loss
        if abs(min_loss-item.loss) <= loss_thres:
            out.append((item.loss, item.results_dict, item.met_tree_num, item.tree_num))
    print("# min loss trees:", len(out))

    return out

######## Methods for comparing Metient outputs to ground truth ############

def evaluate_seeding_clones(sim_clone_tree_fn, sim_vert_labeling_fn, met_results_dict, met_tree_num, met_definition):
    edges_simulated, mig_edges_simulated, _ = parse_clone_tree(sim_clone_tree_fn, sim_vert_labeling_fn)
    if met_definition:
        seeding_clones_simulated = identify_seeding_clones_met(edges_simulated, mig_edges_simulated)
    else: 
        seeding_clones_simulated = identify_seeding_clones_mach(edges_simulated, mig_edges_simulated)
    edges_inferred, mig_edges_inferred = metient_parse_clone_tree(met_results_dict, met_tree_num)
    if met_definition:
        seeding_clones_inferred = identify_seeding_clones_met(edges_inferred, mig_edges_inferred)
    else:
        seeding_clones_inferred = identify_seeding_clones_mach(edges_inferred, mig_edges_inferred)

    # print("edges_inferred", edges_inferred, "seeding_clones_inferred", seeding_clones_inferred)
    # print("edges_simulated", edges_simulated, "seeding_clones_simulated", seeding_clones_simulated)
        
    recall = float(len(seeding_clones_inferred & seeding_clones_simulated)) / float(len(seeding_clones_simulated))
    precision = float(len(seeding_clones_inferred & seeding_clones_simulated)) / float(len(seeding_clones_inferred))
    if recall == 0 or precision == 0:
        F = 0
    else:
        F = 2.0 / ((1.0 / recall) + (1.0 / precision))

    return recall, precision, F

def evaluate_seeding_clones_mach(sim_clone_tree_fn, sim_vert_labeling_fn, met_results_dict, met_tree_num):
    return evaluate_seeding_clones(sim_clone_tree_fn, sim_vert_labeling_fn, met_results_dict, met_tree_num, False)

def evaluate_seeding_clones_met(sim_clone_tree_fn, sim_vert_labeling_fn, met_results_dict, met_tree_num):
    return evaluate_seeding_clones(sim_clone_tree_fn, sim_vert_labeling_fn, met_results_dict, met_tree_num, True)

def evaluate_migration_graph(sim_mig_graph_fn, met_results_dict, met_tree_num):
    edge_set_G_simulated = set(parse_migration_graph(sim_mig_graph_fn))
    edge_set_G_inferred = set(metient_parse_mig_graph(met_results_dict, met_tree_num))
    recall_G = float(len(edge_set_G_inferred & edge_set_G_simulated)) / float(len(edge_set_G_simulated))
    precision_G = float(len(edge_set_G_inferred & edge_set_G_simulated)) / float(len(edge_set_G_inferred))

    if recall_G != 0 and precision_G != 0:
        F_G = 2.0 / ((1.0 / recall_G) + (1.0 / precision_G))
    else:
        F_G = 0

    return recall_G, precision_G, F_G

def evaluate_migration_multigraph(sim_mig_graph_fn, met_results_dict, met_tree_num):
    edge_multiset_G_simulated = multi_graph_to_set(parse_migration_graph(sim_mig_graph_fn))
    edge_multiset_G_inferred = multi_graph_to_set(metient_parse_mig_graph(met_results_dict, met_tree_num))
    recall_G2 = float(len(edge_multiset_G_inferred & edge_multiset_G_simulated)) / float(len(edge_multiset_G_simulated))
    precision_G2 = float(len(edge_multiset_G_inferred & edge_multiset_G_simulated)) / float(len(edge_multiset_G_inferred))
    if recall_G2 != 0 and precision_G2 != 0:
        F_G2 = 2.0 / ((1.0 / recall_G2) + (1.0 / precision_G2))
    else:
        F_G2 = 0

    return recall_G2, precision_G2, F_G2

def is_resolved_polytomy_cluster(cluster_label):
    '''
    In MACHINA simulated data, cluster labels with non-numeric components (e.g. M2_1
    instead of 1;3;4) represent polytomies
    '''
    is_polytomy = False
    for mut in cluster_label.split(";"):
        if not mut.isnumeric() and (mut.startswith('M') or mut.startswith('P')):
            is_polytomy = True
    return is_polytomy
    
    


import numpy as np
import torch
import networkx as nx
import math
from queue import Queue
from metient.util.globals import *
from collections import deque
import gc
import scipy.sparse as sp
import pandas as pd
pd.options.display.float_format = '{:,.3f}'.format
pd.set_option('display.max_columns', None)

LAST_P = None

######################################################
##################### CLASSES ########################
######################################################

class MigrationHistoryNode:
    __slots__ = ['idx', 'label', 'is_witness', 'is_polytomy_resolver_node']
    def __init__(self, idx, label, is_witness=False, is_polytomy_resolver_node=False):
        self.idx = idx
        assert isinstance(label, list)
        label = [str(x) for x in label]
        self.label = label # list of mut names or ids
        self.is_witness = is_witness
        self.is_polytomy_resolver_node = is_polytomy_resolver_node

# Collection of MigrationHistoryNodes which are used for a single migration history
class MigrationHistoryNodeCollection:
    def __init__(self, mig_hist_nodes):
        idx_to_mig_hist_node = {}
        for node in mig_hist_nodes:
            idx_to_mig_hist_node[node.idx] = node
        self.idx_to_mig_hist_node = idx_to_mig_hist_node
    
    @classmethod
    def from_dict(cls, dct):
        nodes = []
        for idx in dct:
            item = dct[idx]
            node = MigrationHistoryNode(idx, item[0], is_witness=item[1], is_polytomy_resolver_node=item[2])
            nodes.append(node)
        instance = cls(nodes)  # Create an instance using the default constructor
        return instance

    def get_nodes(self):
        return list(self.idx_to_mig_hist_node.values())
        
    def idx_to_label(self):
        idx_to_label = {}
        for idx in self.idx_to_mig_hist_node:
            idx_to_label[idx] =  self.idx_to_mig_hist_node[idx].label
        return idx_to_label

    def update_index(self, old_idx, new_idx):

        if old_idx in self.idx_to_mig_hist_node:
            old_node = self.idx_to_mig_hist_node.pop(old_idx)
            old_node.idx = new_idx
            self.idx_to_mig_hist_node[new_idx] = old_node
        else:
            raise KeyError(f"Key '{old_idx}' not found in dictionary.")

    def get_node(self, idx):
        return self.idx_to_mig_hist_node[idx]
    
    def add_node(self, new_node):
        assert new_node.idx not in self.idx_to_mig_hist_node
        self.idx_to_mig_hist_node[new_node.idx] = new_node
    
    def swap_indices(self, key1, key2):
        d = self.idx_to_mig_hist_node
        assert key1 in d and key2 in d
        # Both keys exist, swap their values
        d[key1], d[key2] = d[key2], d[key1]
        d[key1].idx = key1
        d[key2].idx = key2
    
    def remove_indices_and_reindex(self, indices_to_remove):
        # Create a new dictionary to hold the re-indexed entries
        new_dict = {}
        original_dict = self.idx_to_mig_hist_node
        old_index_to_new_index = {}
        # Initialize the new index
        new_index = 0
        # Iterate through the original dictionary in sorted index order
        for old_index in sorted(original_dict.keys()):
            # Skip the indices that need to be removed
            if old_index in indices_to_remove:
                continue
            
            # Assign the new index to the current node
            new_dict[new_index] = original_dict[old_index]
            new_dict[new_index].idx = new_index
            old_index_to_new_index[old_index] = new_index
            # Increment the new index
            new_index += 1
        
        self.idx_to_mig_hist_node = new_dict
        return old_index_to_new_index

# Defines a unique adjacency matrix and vertex labeling
class MigrationHistory:
    def __init__(self, tree, labeling):
        if (tree.shape[0] != tree.shape[1]):
            raise ValueError("Adjacency matrix should have shape (num_nodes x num_nodes)")
        if (tree.shape[0] != labeling.shape[1]):
            raise ValueError("Vertex labeling matrix should have shape (num_sites x num_nodes)")

        self.tree = tree
        self.labeling = labeling

    def _nonzero_tuple(self, tensor):
        # Get the indices of non-zero elements and convert to a hashable form
        if tensor.is_sparse:
            tensor = tensor.coalesce()
            indices = tensor.indices()
        else:
            indices = torch.nonzero(tensor, as_tuple=False)
        return tuple(map(tuple, indices.tolist()))

    def __hash__(self):
        # Compute a hash based on the positions of non-zero entries
        return hash((self._nonzero_tuple(self.labeling), self._nonzero_tuple(self.tree)))

    def __eq__(self, other):
        # Check for equality based on the positions of non-zero entries in both tensors
        if not isinstance(other, MigrationHistory):
            return False
        return (self._nonzero_tuple(self.labeling) == self._nonzero_tuple(other.labeling) and
                self._nonzero_tuple(self.tree) == self._nonzero_tuple(other.tree))

    def __str__(self):
        A = str(torch.where(self.tree != 0))
        V = str(torch.where(self.labeling == 1))
        return f"Tree: {A}\nVertex Labeling: {V}"

def convert_metric_to_int(x):
    if isinstance(x, int):
        return x
    elif isinstance(x, torch.Tensor) and x.numel() == 1:
        return int(x.item())  
    raise ValueError(f"Got unexpected value for metric {x}")

def convert_metric_to_float(x):
    if isinstance(x, float):
        return x
    elif isinstance(x, torch.Tensor) and x.numel() == 1:
        return float(x.item())  
    raise ValueError(f"Got unexpected value for metric {x}")

# Convenience object to package information needed for a final solution
class VertexLabelingSolution:
    def __init__(self, loss, m, c, s, g, o, e, V, soft_V, T, G, node_collection):
        self.loss = loss
        self.m = convert_metric_to_int(m)
        self.c = convert_metric_to_int(c)
        self.s = convert_metric_to_int(s)
        self.g = convert_metric_to_float(g)
        self.o = convert_metric_to_float(o)
        self.e = convert_metric_to_float(e)
        self.V = V
        self.T = T
        self.G = G
        self.node_collection = node_collection
        self.soft_V = soft_V

    # Override the comparison operator
    def __lt__(self, other):
        return self.loss < other.loss
    
    def _nonzero_tuple(self, tensor):
        # Get the indices of non-zero elements and convert to a hashable form
        indices = torch.nonzero(tensor, as_tuple=False)
        return tuple(map(tuple, indices.tolist()))

    def __hash__(self):
        # Compute a hash based on the positions of non-zero entries
        return hash((self._nonzero_tuple(self.V), self._nonzero_tuple(self.T)))

    def __eq__(self, other):
        # Check for equality based on the positions of non-zero entries in both tensors
        if not isinstance(other, VertexLabelingSolution):
            return False
        return (self._nonzero_tuple(self.V) == self._nonzero_tuple(other.V) and
                self._nonzero_tuple(self.T) == self._nonzero_tuple(other.T))

# Convenience object to package information needed for known 
# labelings/indices of vertex labeling matrix V
class FixedVertexLabeling:
    def __init__(self, known_indices, unknown_indices, known_labelings, optimal_root_nodes, old_to_new):
        self.known_indices = known_indices
        self.unknown_indices = unknown_indices
        self.known_labelings = known_labelings
        self.optimal_root_nodes = optimal_root_nodes
        self.old_to_new = old_to_new

######################################################
############# CALCULATING PARSIMONY METRICS ##########
######################################################

def migration_number(site_adj):
    '''
    Args:
        - site_adj: sample_size x num_sites x num_sites matrix, where each num_sites x num_sites
        matrix has the number of migrations from site i to site j
    Returns:
        - migration number: number of total migrations between sites (no same site to same site migrations)
    '''
    site_adj_trace = torch.diagonal(site_adj, offset=0, dim1=1, dim2=2).sum(dim=1)
    m = torch.sum(site_adj, dim=(1, 2)) - site_adj_trace
    return m

def seeding_site_number(site_adj_no_diag):
    '''
    Args:
        - site_adj_no_diag: sample_size x num_sites x num_sites matrix, where each num_sites x num_sites
        matrix has the number of migrations from site i to site j, and no same site migrations are included
    Returns:
        - seeding site number: number of sites that have outgoing edges
    '''
    row_sums_site_adj = torch.sum(site_adj_no_diag, axis=2)
    # can only have a max of 1 for each site (it's either a seeding site or it's not)
    binarized_row_sums_site_adj = torch.sigmoid(BINARY_ALPHA * (2*row_sums_site_adj - 1)) # sigmoid for soft thresholding
    s = torch.sum(binarized_row_sums_site_adj, dim=(1))
    return s

def comigration_number_approximation(site_adj):
    '''
    Args:
        - site_adj: sample_size x num_sites x num_sites matrix, where each num_sites x num_sites
        matrix has the number of migrations from site i to site j
    Returns:
        - comigration number: an approximation of the comigration number that doesn't include a penalty for
        repeated temporal migrations which is expensive to compute
    '''
    binarized_site_adj = torch.sigmoid(BINARY_ALPHA * (2 * site_adj - 1))
    bin_site_trace = torch.diagonal(binarized_site_adj, offset=0, dim1=1, dim2=2).sum(dim=1)
    c = torch.sum(binarized_site_adj, dim=(1,2)) - bin_site_trace
    return c


def comigration_number(site_adj, A, VA, V, VT, update_path_matrix, identical_T):
    '''
    Args:
        - site_adj: sample_size x num_sites x num_sites matrix, where each num_sites x num_sites
        matrix has the number of migrations from site i to site j
        - A: Adjacency matrix (directed) of the full tree (sample_size x num_nodes x num_nodes)
        - VA: V*A (will be converted to sparse)
        - V: Vertex labeling one-hot matrix (sample_size x num_sites x num_nodes)
        - VT: transpose of V 
        - update_path_matrix: whether we need to update the path matrix or we can use a cached version
        (need to update when we're actively resolving polytomies)
        - identical_T: bool, whether all adjacency matrices in T along the sample size dimension are identical
    Returns:
        - comigration number: a subset of the migration edges between two anatomical sites, such that 
        the migration edges occur on distinct branches of the clone tree
    '''

    # First calculate base comigration number (without temporal repeats)
    base_c = comigration_number_approximation(site_adj)
    
    # Get path matrix if needed
    global LAST_P
    if LAST_P is not None and not update_path_matrix:
        P = LAST_P.to(A.device)
        if P.shape != A.shape:
            assert(A.shape[0]==1)
            P = P[0]
    else:
        P = path_matrix(A, remove_self_loops=True, identical_T=identical_T)
        LAST_P = P

    # Get node colors (anatomical site labels) - shape: [batch_size, num_nodes]
    node_colors = torch.argmax(V, dim=1)
    
    # Compute parent colors efficiently using sparse matrix multiplication
    # Shape: [batch_size, num_nodes]
    parent_site_probs = VA.transpose(1, 2)
    parent_colors = torch.argmax(parent_site_probs, dim=2)

    # Find nodes that differ from their parents
    diff_nodes = torch.where(node_colors != parent_colors)  # Returns (batch_indices, node_indices)
    
    temporal_migrations = torch.zeros(node_colors.shape[0], device=V.device)
    
    # Process each batch
    for batch in range(node_colors.shape[0]):
        # Get the differing nodes for this batch
        batch_diff_nodes = diff_nodes[1][diff_nodes[0] == batch]
        
        if len(batch_diff_nodes) > 1:  # Need at least 2 nodes for temporal migrations
            # Create combined key for just these nodes
            diff_node_colors = node_colors[batch, batch_diff_nodes]
            diff_parent_colors = parent_colors[batch, batch_diff_nodes]
            combined_colors = diff_node_colors * V.shape[1] + diff_parent_colors

            # Find groups of nodes with same combination
            unique_combos, inverse_indices, counts = torch.unique(combined_colors, return_inverse=True, return_counts=True)
            
            # Process combinations that appear more than once
            repeated_mask = counts > 1
            repeated_combos = unique_combos[repeated_mask]

             # For each group of matching nodes with same color combination
            for combo in repeated_combos:
                matching_indices = batch_diff_nodes[combined_colors == combo]
                
                # Handle sparse tensor properly
                if P.is_sparse:
                    P_batch = P[batch].coalesce()
                    # Create a mask for the matching indices
                    matching_mask = torch.zeros(P_batch.size(0), dtype=torch.bool, device=P_batch.device)
                    matching_mask[matching_indices] = True
                    
                    # Get all reachable pairs at once
                    source_mask = P_batch.indices()[0][:, None] == matching_indices
                    target_mask = matching_mask[P_batch.indices()[1]]
                    
                    # Count valid paths (avoiding self-paths and double counting)
                    valid_paths = (source_mask & target_mask[:, None]).sum()
                    temporal_migrations[batch] += valid_paths
                else:
                    # For dense tensors, use matrix operations
                    reachability_submatrix = P[batch][matching_indices][:, matching_indices]
                    # Create upper triangular mask to avoid double counting
                    upper_tri_mask = torch.triu(torch.ones_like(reachability_submatrix), diagonal=1, device=reachability_submatrix.device)
                    temporal_migrations[batch] += (reachability_submatrix * upper_tri_mask).sum()


    return base_c + temporal_migrations

    # VAT = torch.transpose(VA, 2, 1)
    # W = VAT @ VA # 1 if two nodes' parents are the same color
    # Y = torch.sum(torch.mul(VAT, 1-VT), axis=2) # Y has a 1 for every node where its parent has a diff color
    # X = VT @ V # 1 if two nodes are the same color
    # shared_par_and_self_color = torch.mul(W, X) # 1 if two nodes' parents are same color AND nodes are same color
    # # tells us if two nodes are (1) in the same site and (2) have parents in the same site
    # # and (3) there's a path from node i to node j
    # global LAST_P # this is expensive to compute, so hash it if we don't need to update it
    # if LAST_P != None and not update_path_matrix:
    #     P = LAST_P.to(A.device)
    #     # Make sure if we're using a cached version (only applicable in cases where 
    #     # we're not resolving polytomies), we provide the correct shape on the first 
    #     # batch dimension. This is only when we're post-processing individual solutions,
    #     # so just grab the first 2-D path matrix
    #     if P.shape != A.shape:
    #         assert(A.shape[0]==1)
    #         P = P[0]
    # else:
    #     P = path_matrix(A, remove_self_loops=True, identical_T=identical_T)
    #     LAST_P = P
    # shared_path_and_par_and_self_color = torch.sum(torch.mul(P, shared_par_and_self_color), axis=2)
    # repeated_temporal_migrations = torch.sum(torch.mul(shared_path_and_par_and_self_color, Y), axis=1)
    # binarized_site_adj = torch.sigmoid(BINARY_ALPHA * (2 * site_adj - 1))
    # bin_site_trace = torch.diagonal(binarized_site_adj, offset=0, dim1=1, dim2=2).sum(dim=1)
    # c = torch.sum(binarized_site_adj, dim=(1,2)) - bin_site_trace + repeated_temporal_migrations
    # return c

def genetic_distance_score(G, m, A, V, VT):
    '''
    Args:
        - G: Matrix of genetic distances between internal nodes (shape: sample_size x num_internal_nodes x num_internal_nodes).
             Lower values indicate lower branch lengths, i.e. more genetically similar.
        - m: vector of migration numbers (length = sample_size)
        - A: Adjacency matrix (directed) of the full tree (sample_size x num_nodes x num_nodes)
        - V: Vertex labeling one-hot matrix (sample_size x num_sites x num_nodes)
        - VT: transpose of V 
    Returns:
        - genetic distance score, for each sample in the first dimension
    '''
    g = torch.zeros(m.shape, device=A.device)
    if G != None:
        X = VT @ V # 1 if two nodes are the same color
        # TODO: is there a way to do this without converting A to a dense matrix?
        A = A.to_dense()
        # Calculate if 2 nodes are in diff sites and there's an edge between them (i.e. there is a migration edge)
        R = torch.mul(A, (1-X))
        adjusted_G = -torch.log(G+0.01)
        R = torch.mul(R, adjusted_G)
        g = torch.sum(R, dim=(1,2))/(m + 1) # to prevent division by 0
    return g

def organotropism_score(O, site_adj_no_diag, p, bs):
    '''
    Args:
        - O: Array of frequencies with which the primary cancer type seeds site i (shape: num_anatomical_sites).
        - site_adj_no_diag: sample_size x num_sites x num_sites matrix, where each num_sites x num_sites
        matrix has the number of migrations from site i to site j, and no self loops
        - p: one-hot vector indicating site of the primary
        - bs: sample_size (number of samples)
    Returns:
        - organotropism score, for each sample in the first dimension
    '''
    o = torch.zeros(bs, device=site_adj_no_diag.device)
    if O != None:
        # the organotropism frequencies can only be used on the first 
        # row, which is for the migrations from primary cancer site to
        # other metastatic sites (we don't have frequencies for every 
        # site to site migration)
        prim_site_idx = torch.nonzero(p)[0][0]
        O = O.repeat(bs,1).reshape(bs, O.shape[0])
        adjusted_freqs = -torch.log(O+0.01)
        num_mig_from_prim = site_adj_no_diag[:,prim_site_idx,:]
        organ_penalty = torch.mul(num_mig_from_prim, adjusted_freqs)
        o = torch.sum(organ_penalty, dim=(1))/(torch.sum(num_mig_from_prim, dim=(1)) + 1) # to prevent division by 0
    return o

def ancestral_labeling_metrics(V, A, G, O, p, update_path_matrix, compute_full_c, identical_T):
    
    single_A = A
    bs = V.shape[0]
    num_sites = V.shape[1]
    A = A if len(A.shape) == 3 else repeat_n(single_A, bs) 
    
    # Compute matrices used for all parsimony metrics
    VA = torch.bmm(A.permute(0,2,1), V.permute(0,2,1)).permute(0,2,1)
    VT = torch.transpose(V, 2, 1)
    site_adj = VA @ VT
    # Remove the same site transitions from the site adjacency matrix
    site_adj_no_diag = torch.mul(site_adj, repeat_n(1-torch.eye(num_sites, num_sites, device=site_adj.device), bs))

    # 1. Migration number
    m = migration_number(site_adj)

    # 2. Seeding site number
    s = seeding_site_number(site_adj_no_diag)

    # 3. Comigration number
    if compute_full_c:
        c = comigration_number(site_adj, A, VA, V, VT, update_path_matrix, identical_T)
    else:
        c = comigration_number_approximation(site_adj)

    # 4. Genetic distance
    g = genetic_distance_score(G, m, A, V, VT)

    # 5. Organotropism
    o = organotropism_score(O, site_adj_no_diag, p, bs)

    return m, c, s, g, o

def calc_entropy(V, soft_V):
    '''
    Args:
        - V: one-hot vertex labeling matrix
        - soft_V: underlying theta parameters of V
    Returns:
        - entropy for the categorical variable representing each node's vertex label
    '''
    eps = 1e-7 # to avoid nans when values in soft_V get very close to 0
    return torch.sum(torch.mul(soft_V, torch.log2(soft_V+eps)), dim=(1, 2))

def get_repeating_weight_vector(bs, weight_list, device):
    # Calculate the number of times each weight should be repeated
    total_weights = len(weight_list)
    repeats = bs // total_weights
    remaining_elements = bs % total_weights

    # Create a list where each weight is repeated 'repeats' times
    repeated_list = [weight for weight in weight_list for _ in range(repeats)]

    # Add remaining elements to match the batch size exactly
    if remaining_elements > 0:
        additional_weights = weight_list[:remaining_elements]
        additional_repeated = [weight for weight in additional_weights]
        repeated_list += additional_repeated

    # Convert the list to a tensor
    weights_vec = torch.tensor(repeated_list, device=device)

    return weights_vec

def get_mig_weight_vector(bs, weights, device):
    return get_repeating_weight_vector(bs, weights.mig, device)

def get_seed_site_weight_vector(bs, weights, device):
    return get_repeating_weight_vector(bs, weights.seed_site, device)

def clone_tree_labeling_loss_with_computed_metrics(m, c, s, g, o, e, weights, bs=1):

    # Combine all 5 components with their weights
    # Explore different weightings
    if isinstance(weights.mig, list) and isinstance(weights.seed_site, list):
        mig_weights_vec = get_mig_weight_vector(bs, weights, m.device)
        seeding_sites_weights_vec = get_seed_site_weight_vector(bs, weights, s.device)
        mig_loss = torch.mul(mig_weights_vec, m)
        seeding_loss = torch.mul(seeding_sites_weights_vec, s)
        labeling_loss = (mig_loss + weights.comig*c + seeding_loss + weights.gen_dist*g + weights.organotrop*o+ weights.entropy*e)
    else:
        mig_loss = weights.mig*m
        seeding_loss = weights.seed_site*s
        labeling_loss = (mig_loss + weights.comig*c + seeding_loss + weights.gen_dist*g + weights.organotrop*o+ weights.entropy*e)
    return labeling_loss

def clone_tree_labeling_objective(V, soft_V, T, G, O, p, weights, 
                                  update_path_matrix=True, compute_full_c=True, identical_T=False):
    '''
    Args:
        V: Vertex labeling of the full tree (sample_size x num_sites x num_nodes)
        T: Adjacency matrix (directed) of the full tree (sample_size x num_nodes x num_nodes)
        G: Matrix of genetic distances between internal nodes (shape: sample_size x num_internal_nodes x num_internal_nodes).
        Lower values indicate lower branch lengths, i.e. more genetically similar.
        O: Array of frequencies with which the primary cancer type seeds site i (shape: num_anatomical_sites).
        p: one-hot vector indicating site of the primary
        weights: Weights object
        update_path_matrix: bool, whether we need to update the path matrix (1:1 with T) or we can use the last
            cached version (this is expensive to compute)
        compute_full_c: bool, whether we need to compute the full comigration number or we can calculate an approzimation
            (this is expensive to compute)
        identical_T: bool, whether all adjacency matrices in T along the sample size dimension are identical

    Returns:
        Loss to score the ancestral vertex labeling of the given tree. This combines (1) migration number, (2) seeding site
        number, (3) comigration number, and optionally (4) genetic distance and (5) organotropism.
    '''
    V = add_batch_dim(V)
    soft_V = add_batch_dim(soft_V)
    m, c, s, g, o = ancestral_labeling_metrics(V, T, G, O, p, update_path_matrix, compute_full_c, identical_T)
    # Entropy
    e = calc_entropy(V, soft_V)

    labeling_loss = clone_tree_labeling_loss_with_computed_metrics(m, c, s, g, o, e, weights, bs=V.shape[0])
    return labeling_loss, (m, c, s, g, o, e)

def stack_fixed_labeling(X, fixed_labeling, p):
    """
    Helper function to stack vertex labeling with fixed labelings.
    Handles both cases where old_to_new mapping is present or None.
    
    Args:
        X: Tensor of shape (bs, num_sites, num_unknown_nodes) containing unknown labelings
        fixed_labeling: FixedVertexLabeling object containing known indices and labelings
        p: Primary site labeling tensor
    
    Returns:
        full_X: Tensor containing both known and unknown labelings properly mapped
    """
    bs = X.shape[0]
    # Initialize with space for p + known/unknown nodes
    
    known_labelings = repeat_n(fixed_labeling.known_labelings, bs)
    p = repeat_n(p, bs).squeeze(-1)
    
    if fixed_labeling.old_to_new is None:
        full_X = torch.zeros((bs, X.shape[1], 1 + len(fixed_labeling.known_indices)+len(fixed_labeling.unknown_indices)))
        # Set primary site labeling first
        full_X[:, :, 0] = p
        # Simple case: direct mapping of indices
        full_X[:, :, fixed_labeling.unknown_indices] = X
        full_X[:, :, fixed_labeling.known_indices] = known_labelings
    else:
        # Complex case: use old_to_new mapping
        full_X = torch.zeros((bs, X.shape[1], 1 + len(fixed_labeling.optimal_root_nodes) + len(fixed_labeling.unknown_indices)))
        # Set primary site labeling first
        full_X[:, :, 0] = p
        root_indices = [idx for idx in fixed_labeling.known_indices if idx in fixed_labeling.optimal_root_nodes]
        root_src_idx = torch.tensor([fixed_labeling.known_indices.index(idx) for idx in root_indices])
        root_dst_idx = torch.tensor([fixed_labeling.old_to_new.get(idx, idx) for idx in root_indices])
        
        unknown_indices = fixed_labeling.unknown_indices 
        unknown_src_idx = torch.arange(len(unknown_indices))
        unknown_dst_idx = torch.tensor([fixed_labeling.old_to_new.get(idx, idx) for idx in unknown_indices], dtype=torch.long)
        
        full_X[:, :, root_dst_idx] = fixed_labeling.known_labelings[:, root_src_idx].unsqueeze(0).expand(X.shape[0], -1, -1)
        full_X[:, :, unknown_dst_idx] = X[:, :, unknown_src_idx]
    
    return full_X

def stack_vertex_labeling(L, X, p, poly_res, fixed_labeling):
    '''
    Use leaf labeling L and X (both of size sample_size x num_sites X num_internal_nodes)
    to get the anatomical sites of the leaf nodes and the internal nodes (respectively). 
    Stack the root labeling to get the full vertex labeling V. 
    '''
    # Expand leaf node labeling L to be repeated sample_size times
    bs = X.shape[0]
    L = repeat_n(L, bs)

    if fixed_labeling != None:
        full_X = stack_fixed_labeling(X, fixed_labeling, p)
    else:
        # Include p at the start for the non-fixed labeling case
        p = repeat_n(p, bs)
        full_X = torch.cat((p, X), dim=2)

    # Concatenate with leaf labelings
    # Order is: internal nodes, new poly nodes (is resolving polytomies), leaf nodes from U
    full_vert_labeling = torch.cat((full_X, L), dim=2)
    return full_vert_labeling

######################################################
######### POST U MATRIX ESTIMATION UTILITIES #########
######################################################

def get_leaf_labels_from_U(U):
    
    U = U[:,1:] # don't include column for normal cells
    internal_node_idx_to_sites = {}
    for node_idx in range(U.shape[1]):
        for site_idx in range(U.shape[0]):
            node_U = U[site_idx,node_idx]
            if node_U > U_CUTOFF:
                if node_idx not in internal_node_idx_to_sites:
                    internal_node_idx_to_sites[node_idx] = []
                internal_node_idx_to_sites[node_idx].append(site_idx)
    
    return internal_node_idx_to_sites


def full_adj_matrix_from_internal_node_idx_to_sites_present(input_T, input_G, idx_to_sites_present, 
                                                            num_sites, G_identical_clone_val, 
                                                            node_collection, ordered_sites):
    '''
    All non-zero values of U represent extant clones (witness nodes of the full tree).
    For each of these non-zero values, we add an edge from parent clone to extant clone.
    '''
    num_leaves = sum(len(lst) for lst in idx_to_sites_present.values())
    full_adj = torch.nn.functional.pad(input=input_T, pad=(0, num_leaves, 0, num_leaves), mode='constant', value=0)
    witness_idx = input_T.shape[0]
    # Also add branch lengths (genetic distances) for the edges we're adding.
    # Since all of these edges we're adding represent genetically identical clones,
    # we are going to add a very small but non-zero value.
    full_G = torch.nn.functional.pad(input=input_G, pad=(0, num_leaves, 0, num_leaves), mode='constant', value=0) if input_G is not None else None

    leaf_labels = []
    # Iterate through the internal nodes that we want to attach witness nodes to
    for internal_node_idx in idx_to_sites_present:
        # Attach a leaf node for every site that this internal node is observed in
        for site in idx_to_sites_present[internal_node_idx]:
            full_adj[internal_node_idx, witness_idx] = 1
            label = node_collection.get_node(internal_node_idx).label + [ordered_sites[site]]
            leaf_node = MigrationHistoryNode(witness_idx, label, is_witness=True, is_polytomy_resolver_node=False)
            node_collection.add_node(leaf_node)
            if input_G is not None:
                full_G[internal_node_idx, witness_idx] = G_identical_clone_val
            witness_idx += 1
            leaf_labels.append(site)

    assert len(leaf_labels) !=0, f"No nodes present in any sites"

    # Anatomical site labels of the leaves
    L = torch.nn.functional.one_hot(torch.tensor(leaf_labels), num_classes=num_sites).T
    return full_adj, full_G, L, node_collection
    
def full_adj_matrix_using_inferred_observed_clones(U, input_T, input_G, num_sites, G_identical_clone_val,
                                                   node_collection, ordered_sites):
    '''
    Use inferred observed clones to fill out T and G by adding leaf nodes
    '''
    internal_node_idx_to_sites = get_leaf_labels_from_U(U)
    full_adj, full_G, L, node_collection = full_adj_matrix_from_internal_node_idx_to_sites_present(input_T, input_G, internal_node_idx_to_sites, 
                                                                                                   num_sites, G_identical_clone_val,
                                                                                                   node_collection, ordered_sites)
    
    return full_adj, full_G, L, internal_node_idx_to_sites, node_collection

def remove_leaf_indices_not_observed_sites(removal_indices, U, input_T, T, G, node_collection, idx_to_observed_sites):
    '''
    Remove clone tree leaf nodes that are not detected in any sites. 
    These are not well estimated
    '''
    if len(removal_indices) == 0:
        return U, input_T, T, G, node_collection, idx_to_observed_sites
    
    for remove_idx in removal_indices:
        child_indices = get_child_indices(T, [remove_idx])
        assert len(child_indices) == 0

    # Remove indices from input_T, full T (now with observed clones), U and G
    T = np.delete(T, removal_indices, 0)
    T = np.delete(T, removal_indices, 1)

    # Remove indices from T, U and G
    input_T = np.delete(input_T, removal_indices, 0)
    input_T = np.delete(input_T, removal_indices, 1)

    U = np.delete(U, removal_indices, 1)

    if G != None: 
        G = np.delete(G, removal_indices, 0)
        G = np.delete(G, removal_indices, 1)

    # Reindex the idx to label dict
    old_index_to_new_index = node_collection.remove_indices_and_reindex(removal_indices)
    new_idx_to_observed_sites = {}
    for old_idx in idx_to_observed_sites:
        new_idx = old_index_to_new_index[old_idx]
        new_idx_to_observed_sites[new_idx] = idx_to_observed_sites[old_idx]

    return U, input_T, T, G, node_collection, new_idx_to_observed_sites

######################################################
################## RANDOM UTILITIES ##################
######################################################

def multiply_with_sparse_check(m1, m2):
    """
    Multiplies a dense matrix m1 with a sparse matrix m2
    
    If m2 is a sparse tensor, it uses sparse-dense multiplication.
    If m2 is a dense tensor, it uses standard dense multiplication.
    """
    if isinstance(m2, torch.sparse.SparseTensor):
        # If m2 is sparse, multiply using sparse-dense multiplication
        return torch.sparse.mm(m2.transpose(0, 1), V.T).T
    else:
        # If m2 is dense, multiply using standard multiplication
        return m1 @ m2
    
def print_gpu_memory():
    if torch.cuda.is_available():
        # Get current device index
        device = torch.cuda.current_device()

        # Print the name of the GPU
        print(f"Using device: {torch.cuda.get_device_name(device)}")
        
        # Total memory
        total_memory = torch.cuda.get_device_properties(device).total_memory / 1e9  # in GB
        
        # Reserved memory by tensors
        reserved_memory = torch.cuda.memory_reserved(device) / 1e9  # in GB
        
        # Memory actually being used by tensors
        allocated_memory = torch.cuda.memory_allocated(device) / 1e9  # in GB
        
        # Available free memory
        free_memory = reserved_memory - allocated_memory
        
        print(f"Total memory: {total_memory:.2f} GB")
        print(f"Reserved memory: {reserved_memory:.2f} GB")
        print(f"Allocated memory: {allocated_memory:.2f} GB")
        print(f"Free memory: {free_memory:.2f} GB\n")
    else:
        print("CUDA is not available")


def mutation_matrix_with_normal_cells(T):
    B = mutation_matrix(T)
    # Add a row of zeros to account for the non-cancerous root node
    B = torch.vstack([torch.zeros(B.shape[1]), B])
    # Add a column of ones to indicate that every clone is a descendent of the non-cancerous root node
    B = torch.hstack([torch.ones(B.shape[0]).reshape(-1,1), B])
    return B

def adjacency_matrix_to_edge_list(adj_matrix):
    edges = []
    for i, j in tree_iterator(adj_matrix):
        edges.append((i, j))
    return edges

def reindex_dict(original_dict, indices_to_remove):
    # Create a new dictionary to hold the re-indexed entries
    new_dict = {}
    
    old_index_to_new_index = {}
    # Initialize the new index
    new_index = 0
    # Iterate through the original dictionary in sorted index order
    for old_index in sorted(original_dict.keys()):
        # Skip the indices that need to be removed
        if old_index in indices_to_remove:
            continue
        
        # Assign the new index to the current label
        new_dict[new_index] = original_dict[old_index]
        old_index_to_new_index[old_index] = new_index
        # Increment the new index
        new_index += 1
    
    return new_dict, old_index_to_new_index

def add_batch_dim(x):
    if len(x.shape) == 3:
        return x
    return x.reshape(1, x.shape[0], x.shape[1])

def to_tensor(t):
    if torch.is_tensor(t):
        return t
    return torch.stack(t)

def create_reweighted_solution_set_from_pckl(pckl, O, p, weights):
    # Make a solution set from the pickled files
    Ts, Vs, soft_Vs, Gs = pckl[OUT_ADJ_KEY], pckl[OUT_LABElING_KEY], pckl[OUT_SOFTV_KEY], pckl[OUT_GEN_DIST_KEY]
    node_collections = [MigrationHistoryNodeCollection.from_dict(dct) for dct in pckl[OUT_IDX_LABEL_KEY]]
    solution_set = []
    for T, V, soft_V, G, node_collection in zip(Ts, Vs, soft_Vs, Gs, node_collections):
        T = torch.sparse_coo_tensor(T[0], T[1], size=T[2])
        V,soft_V,G = torch.tensor(V),torch.tensor(soft_V),torch.tensor(G)
        loss, new_metrics = clone_tree_labeling_objective(V,soft_V, T, G, O, p, weights, True)
        solution_set.append(VertexLabelingSolution(loss, *new_metrics,V,soft_V, T, G,node_collection))
    return solution_set

def calculate_sample_size(num_nodes, num_sites, solve_polytomies):
    '''
    Calculate the number of samples to initialize for a run based on 
    the number of tree nodes, the number of anatomical sites, and if we're
    solving polytomies
    '''
    min_size = 256
    min_size += num_nodes*num_sites*4

    if solve_polytomies:
        min_size *= 2

    # cap this to a reasonably high sample size
    min_size = min(min_size, 60000)
    return min_size

def tree_iterator(T):
    ''' 
    Iterate an adjacency matrix, yielding i and j for all values = 1.
    Handles dense NumPy arrays, dense PyTorch tensors, and sparse PyTorch tensors.
    '''
    # Case 1: If input is a sparse PyTorch tensor
    if isinstance(T, torch.Tensor) and T.is_sparse:
        non_zero_indices = T.coalesce().indices()  # Get non-zero indices from sparse tensor
        
        # Yield i, j for all non-zero values (equal to 1 in adjacency matrix)
        for i in range(non_zero_indices.size(1)):
            row, col = non_zero_indices[:, i]
            yield int(row), int(col)

    # Case 2: If input is a dense PyTorch tensor
    elif isinstance(T, torch.Tensor):
        T = T.detach().cpu()  # Ensure it's on the CPU
        non_zero_indices = torch.nonzero(T != 0, as_tuple=False)  # Get indices where value == 1

        # Yield i, j for all non-zero values
        for row, col in non_zero_indices:
            yield int(row), int(col)

    # Case 3: If input is a NumPy array
    elif isinstance(T, np.ndarray):
        # Convert NumPy array to a SciPy sparse matrix for efficiency
        T_sparse = sp.coo_matrix(T)  # Convert to sparse COO format

        # Yield i, j for all non-zero values
        for row, col in zip(T_sparse.row, T_sparse.col):
            yield int(row), int(col)

    else:
        raise TypeError("Input must be a sparse PyTorch tensor, dense PyTorch tensor, or NumPy array.")


def list_gpu_tensors():
    for obj in gc.get_objects():
        try:
            if torch.is_tensor(obj) or (hasattr(obj, 'data') and torch.is_tensor(obj.data)):
                if obj.is_cuda:  # Check if the tensor is on the GPU
                    print(f"Tensor on GPU - shape: {obj.shape}, dtype: {obj.dtype}")
        except Exception as e:
            pass

def bfs_iterator(tree, start_node):
    '''
    Iterate an adjacency matrix in breadth first search order
    '''
    queue = deque([start_node])  # Start the queue with the given start node index
    visited = set(queue)         # Mark the start node as visited

    # BFS loop that yields nodes in the order they are visited
    while queue:
        current = queue.popleft()
        yield current  # Yield the current node

        # Check each adjacency in the row for the current node
        for neighbor, is_connected in enumerate(tree[current]):
            if is_connected and neighbor not in visited:
                visited.add(neighbor)
                queue.append(neighbor)

def has_leaf_node(adjacency_matrix, start_node, num_internal_nodes):
    num_nodes = adjacency_matrix.size(0)
    visited = torch.zeros(num_nodes, dtype=torch.bool)
    queue = Queue()

    queue.put(start_node)
    visited[start_node] = True

    while not queue.empty():
        current_node = int(queue.get())
        # Get neighbors of the current node
        neighbors = torch.nonzero(adjacency_matrix[current_node]).squeeze(1)

        # Enqueue unvisited neighbors
        for neighbor in neighbors:
            neighbor = int(neighbor)
            if not visited[neighbor]:
                queue.put(neighbor)
                visited[neighbor] = True
                # it's a leaf node if its not in the clone tree idx label dict
                if neighbor > num_internal_nodes:
                    return True
    return False

def get_root_index(T):
    '''
    Returns the root idx (node with no inbound edges) from adjacency matrix T
    '''

    candidates = set([x for x in range(len(T))])
    for _, j in tree_iterator(T):
        candidates.remove(j)
    msg = "More than one" if len(candidates) > 1 else "No"
    assert (len(candidates) == 1), f"{msg} root node detected"

    return list(candidates)[0]

def reverse_bfs_order(A):
    '''
    Returns nodes in reverse bfs order of adjacency matrix A
    '''
    root_idx = get_root_index(A)
    nodes = []
    for x in bfs_iterator(A, root_idx):
        nodes.append(x)
    nodes.reverse()
    return nodes

def get_leaves(A):
    '''
    Returns leaves of adjacency matrix A
    '''
    return [int(x) for x in torch.where(A.sum(dim=1)==0)[0]]

def swap_keys(d, key1, key2):
    if key1 in d and key2 in d:
        # Both keys exist, swap their values
        d[key1], d[key2] = d[key2], d[key1]
    elif key1 in d and key2 not in d:
        # Only key1 exists, move its value to key2
        d[key2] = d.pop(key1)
    elif key2 in d and key1 not in d:
        # Only key2 exists, move its value to key1
        d[key1] = d.pop(key2)
    # If neither key exists, do nothing
    return d

def find_leaf_nodes(sparse_tensor):
    """
    Given a sparse COO tensor representing an adjacency matrix, find the leaf nodes.
    
    Parameters:
    - sparse_tensor: Sparse COO tensor representing the adjacency matrix (n x n).
    
    Returns:
    - leaf_nodes: A list of indices of leaf nodes (nodes with no outgoing edges).
    """
    # Get the "from" nodes (row indices) of the sparse matrix
    from_nodes = sparse_tensor.coalesce().indices()[0]

    # Identify all nodes in the graph (based on the number of rows/columns in the matrix)
    all_nodes = torch.arange(sparse_tensor.shape[0])

    # Leaf nodes are those that are not present in the "from" node list
    leaf_nodes = all_nodes[~torch.isin(all_nodes, from_nodes)].tolist()

    return leaf_nodes

def restructure_matrices_root_index_zero(adj_matrix, ref_matrix, var_matrix, node_collection, gen_dist_matrix, idx_to_observed_sites):
    '''
    Restructure the inputs so that the node at index 0 becomes the root node.
    '''
    og_root_idx = get_root_index(adj_matrix)
    return restructure_matrices(og_root_idx, 0, adj_matrix, ref_matrix, var_matrix, 
                                node_collection, gen_dist_matrix, idx_to_observed_sites, None, None)

def restructure_matrices(source_root_idx, target_root_idx, adj_matrix, ref_matrix, var_matrix, 
                         node_collection, gen_dist_matrix, idx_to_observed_sites, V, U):
    '''
    Restructure the inputs so that the order of nodes has source_root_idx and target_root_idx swapped

    Returns:
        - Restructured adjacency matrix, reference matrix, variant matrix, node_idt_to_label dcitionary,
        and genetic distance matrix
    '''
    if source_root_idx == target_root_idx:
        # Nothing to restructure here!
        return adj_matrix, ref_matrix, var_matrix, node_collection, gen_dist_matrix, idx_to_observed_sites, V, U
    
    new_order = [x for x in range(len(adj_matrix))]
    new_order[source_root_idx] = target_root_idx
    new_order[target_root_idx] = source_root_idx
    # Swap rows and columns to move the first row and column to the desired position
    swapped_adjacency_matrix = adj_matrix[new_order, :][:, new_order]

    # Swap columns 
    swapped_ref_matrix = ref_matrix[:, new_order] if ref_matrix != None else None
    swapped_var_matrix = var_matrix[:, new_order] if var_matrix != None else None
    swapped_V = V[:, new_order] if V != None else None
    
    if U == None:
        swapped_U = None
    else:
        normal_cells = U[:,0].view(-1, 1)
        U_order = [x+1 for x in new_order if x < U.shape[1]-1]
        swapped_U = U[:, U_order]
        swapped_U = torch.cat((normal_cells, swapped_U), dim=1)
        
    if gen_dist_matrix == None:
        swapped_gen_dist_matrix = None
    else:
        swapped_gen_dist_matrix = gen_dist_matrix[new_order, :][:, new_order]
    
    if idx_to_observed_sites == None:
        idx_to_observed_sites = None
    else:
        idx_to_observed_sites = swap_keys(idx_to_observed_sites, source_root_idx, target_root_idx)

    node_collection.swap_indices(source_root_idx, target_root_idx)
    
    return swapped_adjacency_matrix, swapped_ref_matrix, swapped_var_matrix, node_collection, swapped_gen_dist_matrix, idx_to_observed_sites, swapped_V, swapped_U

def nodes_w_leaf_nodes(adj_matrices, num_internal_nodes):
    '''
    Args:
        - adj_matrices: 3d matrix, where each inner matrix is a 2d adjacency matric
        - num_internal_nodes: number of nodes in the clone tree which are not leaf nodes
          indicating clone presences

    Returns:
        A list of lists, where list i is for adjacency matrix i, and the inner list is a list
        of booleans indicated whether node j has a leaf clone presence node or not
    '''
    # Convert the 3D adjacency matrix to a PyTorch tensor
    adj_tensor = torch.tensor(adj_matrices, dtype=torch.float32)
    
    # Get dimensions
    _, n, _ = adj_tensor.size()

    # Step 2: Create a mask for columns with indices >= num_internal_nodes
    col_indices = torch.arange(n)
    col_mask = col_indices >= num_internal_nodes

    # Step 3: Apply the column mask to the 3D adjacency matrix
    # Broadcast the col_mask to the shape of the last dimension of adj_tensor
    filtered_cols = adj_tensor[:, :, col_mask]

    # Step 4: Check for each row in each 2D matrix if there's at least one '1' in the filtered columns
    mask = filtered_cols.any(dim=2)

    return mask

def print_U(U, B, node_collection, ordered_sites, ref, var):
    cols = ["GL"]+[";".join([str(i)]+node_collection.get_node(i).label[:2]) for i in range(len(node_collection.get_nodes())) if not node_collection.get_node(i).is_witness]
    U_df = pd.DataFrame(U.detach().numpy(), index=ordered_sites, columns=cols)

    print("U\n", U_df)
    F_df = pd.DataFrame((var/(ref+var)).numpy(), index=ordered_sites, columns=cols[1:])
    print("F\n", F_df)
    Fhat_df = pd.DataFrame(0.5*(U @ B).detach().numpy()[:,1:], index=ordered_sites, columns=cols[1:])
    print("F hat\n", Fhat_df)

def top_k_integers_by_count(lst, k, min_num_sites, cutoff):
    # find unique sites and their counts
    unique_sites, counts = np.unique(lst, return_counts=True)
    # filter for sites that occur at least min_num_sites times
    filtered_values = unique_sites[counts >= min_num_sites]
    # recount after filtering
    unique_sites, counts = np.unique(filtered_values, return_counts=True)
    # sort based on counts in descending order
    sorted_indices = np.argsort(counts)[::-1]
    # get the top k integers and their counts
    if len(unique_sites) > k and cutoff:
        return list(unique_sites[sorted_indices[:k]])
        
    return list(unique_sites)

def traverse_until_Uleaf_or_mult_children(node, input_T, internal_node_idx_to_sites):
    """
    Traverse down the tree on a linear branch until a leaf node is found, and then
    return that node
    
    Special case where a node isn't detected in any sites and is on a long linear branch,
    so there are no children or grandchildren observed to help bias towards
    """
    if node in internal_node_idx_to_sites:
        return node
    
    children = input_T[node].nonzero()[:,0].tolist()
    
    if len(children) == 1:
        return traverse_until_Uleaf_or_mult_children(children[0], input_T, internal_node_idx_to_sites)
    
    return None  # no node has a leaf or node has more than one child
   
def get_resolver_sites_of_node_and_children(input_T, node_idx_to_sites, num_children, node_idx, include_children, min_num_sites, cutoff):
    '''
    Looking at the sites that node_idx and node_idx's children are in,
    figure out the sites that the polytomy resolver nodes should be initialized with
    (any sites that are detected >= thres times)
    '''
   
    clone_tree_children = input_T[node_idx].nonzero()[:,0].tolist()
    # calculated based on clone tree children and leaf nodes estimated form U
    num_possible_resolvers = math.floor(num_children/2)
    # get the sites that the node and node's children are in
    if include_children:
        node_and_children = clone_tree_children+ [int(node_idx)]
    else:
        node_and_children = [int(node_idx)]
    sites = []
    for key in node_and_children:
        if key in node_idx_to_sites:
            sites.extend(node_idx_to_sites[key])
    # print("node_idx", node_idx, "node_and_children", node_and_children, "sites", sites)
    # Special case where the node of interest (node_idx) isn't detected in any sites
    # and is on a linear branch 
    if not include_children and len(sites) == 0 and len(clone_tree_children) == 1:
        leaf_on_linear_branch = traverse_until_Uleaf_or_mult_children(node_idx, input_T, node_idx_to_sites)
        # print("leaf_on_linear_branch", leaf_on_linear_branch)
        if leaf_on_linear_branch != None:
            sites = node_idx_to_sites[leaf_on_linear_branch]
    
    return top_k_integers_by_count(sites, num_possible_resolvers, min_num_sites, cutoff)
    
def get_k_or_more_children_nodes(input_T, T, internal_node_idx_to_sites, k, include_children, min_num_sites, cutoff=True):
    '''
    returns the indices and proposed labeling for nodes
    that are under nodes with k or more children
    e.g. 
    input_T =[[0,1,1,],
             [0,0,0],
             [0,0,0],]
    T = [[0,1,1,1,0,0,0],
         [0,0,0,0,1,0,1],
         [0,0,0,0,0,1,0],
         [0,0,0,0,0,0,0],
         [0,0,0,0,0,0,0,0],
         [0,0,0,0,0,0,0,0],
         [0,0,0,0,0,0,0,0],]
    internal_node_idx_to_sites = {0:[1], 1:[0,1], 2:[1}
    k = 3
    include_children = True
    min_num_sites = 1
    returns ([0], [[1]]) (node at index 0 has 3 children, and the resolver node under it should be
    place in site 1, since node 0 and its child node 1 are both detected in site 1)

    if include_children is False, we only look at node 0's U values and not its children's U values
    '''
    
    T_dense = T.to_dense() if T.is_sparse else T
    row_sums = torch.sum(T_dense, axis=1)
    node_indices = list(torch.where(row_sums >= k)[0])
    del T_dense
    filtered_node_indices = []
    all_resolver_sites = []
    for node_idx in node_indices:
        # get the sites that node_idx and node_idx's children are estimated in U
        resolver_sites = get_resolver_sites_of_node_and_children(input_T, internal_node_idx_to_sites, row_sums[node_idx], node_idx, include_children, min_num_sites, cutoff)
        # print("resolver_sites", resolver_sites)
        # a resolver node wouldn't help for this polytomy
        if len(resolver_sites) == 0:
            continue
        filtered_node_indices.append(int(node_idx))
        all_resolver_sites.append(resolver_sites)
    return filtered_node_indices, all_resolver_sites

def find_first_branching_point(adj_matrix):
    out_degrees = adj_matrix.sum(dim=1)  # Calculate out-degrees of each node
    for node, out_degree in enumerate(out_degrees):
        if out_degree > 1:
            return node
    return None  # Return None if no branching point is found

import torch
import networkx as nx
from itertools import product

# Helper function to build a NetworkX tree from an adjacency matrix
def build_tree(adj_matrix):
    G = nx.DiGraph()
    n = adj_matrix.shape[0]
    for i in range(n):
        for j in range(n):
            if adj_matrix[i, j] == 1:
                G.add_edge(i, j)
    return G

# Helper function to find the first common branching point in two graphs
def find_first_branching_point(adj_matrix):
    out_degrees = adj_matrix.sum(dim=1)  # Calculate out-degrees of each node
    for node, out_degree in enumerate(out_degrees):
        if out_degree > 1:
            return node
    return None  # Return None if no branching point is found

# Helper function to extract subtrees rooted at the given node
def extract_subtrees(G, root):
    subtrees = []
    for child in G.successors(root):
        subtree = nx.dfs_tree(G, source=child)
        subtrees.append(subtree)
    return subtrees

# Helper function to combine subtrees and labels from multiple solutions
def combine_subtrees(subtrees_list, labels_list, common_root):
    combinations = []
    num_subtrees = len(subtrees_list[0])
    
    # Generate all combinations of subtrees from each solution
    for indices in product(range(num_subtrees), repeat=len(subtrees_list)):
        combined_adj_matrix = torch.zeros_like(subtrees_list[0][0].adj_matrix)
        combined_labels = torch.zeros_like(labels_list[0])
        
        for i, idx in enumerate(indices):
            subtree = subtrees_list[i][idx]
            labels = labels_list[i]
            # Combine adjacency matrices
            for edge in subtree.edges:
                combined_adj_matrix[edge[0], edge[1]] = 1
            # Combine labels
            for node in subtree.nodes:
                combined_labels[:, node] = labels[:, node]
        
        combinations.append((combined_adj_matrix, combined_labels))
    return combinations

def get_child_indices(T, indices):
    '''
    returns the indices of direct children of the nodes at indices
    '''
    all_child_indices = []

    for parent_idx in indices:
        children = torch.where(T[parent_idx,:] > 0)[0]
        for child_idx in children:
            if child_idx not in all_child_indices:
                all_child_indices.append(int(child_idx))

    return all_child_indices

def get_child_indices_sparse_t(T, optimal_batch_num, parent_idx):
    T = T.coalesce()
    indices = T.indices()
    bss,iss,jss = indices[0], indices[1], indices[2]
    child_indices = []
    for b,i,j in zip(bss,iss,jss):
        if b == optimal_batch_num and i == parent_idx:
            child_indices.append(int(j))
    return child_indices

def get_parent(T, node):
    num_nodes = len(T)
    parents = [i for i in range(num_nodes) if T[i][node] != 0]
    assert len(parents) == 1
    return parents[0]

def find_parents_children(T, node):
    num_nodes = len(T)
    
    # Find parents: Look in the node's column
    parents = [i for i in range(num_nodes) if T[i][node] != 0]
    
    # Find children: Look in the node's row
    children = [j for j in range(num_nodes) if T[node][j] != 0]
    
    return parents, children

def repeat_n(x, n):
    '''
    Repeats tensor x 'n' times along the first axis, returning a tensor
    w/ dim (n, x.shape[0], x.shape[1])
    '''
    if n == 0:
        return x
    if x.is_sparse:
        return repeat_sparse(x, n)
    return x.repeat(n,1).reshape(n, x.shape[0], x.shape[1])

def repeat_sparse(x, n):
    if not x.is_coalesced():
        x = x.coalesce()
        
    # x: the input sparse tensor
    indices = x.indices()  # Get the indices of the sparse tensor
    values = x.values()    # Get the values of the sparse tensor
    size = x.size()        # Get the size of the sparse tensor

    # Create batch indices for all repetitions (n)
    batch_indices = torch.arange(n).repeat_interleave(indices.shape[1]).to(x.device)
    
    # Expand original indices to match the batch size (adding a new dimension)
    expanded_indices = indices.unsqueeze(0).repeat(n, 1, 1)  # Shape: (n, num_dims, num_non_zero)
    expanded_indices = expanded_indices.permute(1, 0, 2).reshape(expanded_indices.shape[1], expanded_indices.shape[0] * expanded_indices.shape[2])

    # Concatenate batch indices to expanded original indices
    final_indices = torch.cat([batch_indices.unsqueeze(0),expanded_indices], dim=0)
    # Repeat the values across the batch dimension
    repeated_values = values.repeat(n)

    # Define the new size, which includes the batch dimension
    new_size = torch.Size([n, *size])
    
    # Create a new sparse tensor with batch-wise repeated indices and values
    return torch.sparse_coo_tensor(final_indices, repeated_values, new_size)


def stack_closures(closures, batch_size, n):
    # Stack the closures into a single sparse tensor
    stacked_indices = []
    stacked_values = []

    for i in range(batch_size):
        closure = closures[i]
        indices = closure.indices().clone()
        values = closure.values()

        # Adjust row indices for stacking
        indices[0] += i * n

        stacked_indices.append(indices)
        stacked_values.append(values)

    # Concatenate indices and values
    final_indices = torch.cat(stacked_indices, dim=1)
    final_values = torch.cat(stacked_values)

    # Create the final stacked sparse tensor
    stacked_closure = torch.sparse_coo_tensor(final_indices, final_values, (batch_size, n, n))

    return stacked_closure

def topological_sort_dag(T):
    """
    Perform topological sorting on the DAG using NetworkX.
    """
    # Convert adjacency matrix to NetworkX DiGraph
    G = nx.DiGraph(T.cpu().numpy())
    
    # Return the nodes in topological sorted order
    topo_sort = list(nx.topological_sort(G))
    topo_sort.reverse()
    return topo_sort

def purdom_transitive_closure(T, remove_self_loops):
    """
    Optimized version of Purdom's algorithm to compute transitive closure 
    focusing only on outgoing edges.
    
    T: Sparse adjacency matrix of the DAG.
    Returns: Sparse binary reachability matrix
    """
    T = T.to_dense().byte()
    
    # Initialize reachability matrix in-place
    reachability_matrix = T.clone()

    # Perform topological sort of the DAG
    topo_sort = topological_sort_dag(T)
    # Traverse the nodes in topological order
    for u in topo_sort:
        # Get the outgoing neighbors of node u
        neighbors_u = (T[u] != 0).nonzero(as_tuple=False)
        neighbors_u = [idx.item() for idx in neighbors_u]

        if neighbors_u:
            for v in neighbors_u:
                # Propagate reachability for the outgoing edges from u
                reachability_matrix[u] |= reachability_matrix[v]
    
    if not remove_self_loops:
        diag_ones = torch.eye(T.shape[0])
        # Add the diagonal ones to the original matrix
        reachability_matrix = reachability_matrix + diag_ones

    return reachability_matrix.to_sparse().coalesce()

def path_matrix(T, remove_self_loops=False, identical_T=False):
    '''
    Compute the transitive closure of adjacency matrix T using Purdom's algorithm
    '''
    
    # If there's a single 2D matrix, compute its closure directly
    if len(T.shape) == 2:
        return purdom_transitive_closure(T, remove_self_loops)
    
    # For batched matrices
    closures = []
    og_bs = T.shape[0]
    bs = 1 if identical_T else og_bs
    for i in range(bs):
        P = purdom_transitive_closure(T[i], remove_self_loops)
        closures.append(P)

    if identical_T:
        stacked_closure = repeat_n(closures[0], og_bs)
    else:
        stacked_closure = torch.stack(closures)
    return stacked_closure

def mutation_matrix(A):
    '''
    A is an numpy ndarray or tensor adjacency matrix (where Aij = 1 if there is a path from i to j)

    returns a mutation matrix B, which is a subclone x mutation binary matrix, where Bij = 1
    if subclone i has mutation j.
    '''
    return path_matrix(A.T, remove_self_loops=False)

def get_adj_matrix_from_edge_list(edge_list):
    T = []
    nodes = set([node for edge in edge_list for node in edge])
    T = [[0 for _ in range(len(nodes))] for _ in range(len(nodes))]
    for edge in edge_list:
        T[ord(edge[0]) - 65][ord(edge[1]) - 65] = 1
    return torch.tensor(T, dtype = torch.float32)

def is_tree(adj_matrix):
    # rows, cols = torch.where(adj_matrix == 1)
    # edges = zip(rows.tolist(), cols.tolist())
    edges = adjacency_matrix_to_edge_list(adj_matrix)
    g = nx.Graph()
    g.add_edges_from(edges)
    return (not nx.is_empty(g) and nx.is_tree(g))

def pareto_front(solutions, all_pars_metrics):
    """ 
    Args:
        - solutions: list of VertexLabelingSolutions of length n
        - all_pars_metrics: list of tuples with parsimony mentrics
        for (migration #, comigration #, seeding site #) of length n

    Returns:
        - a list of Pareto optimal parsimony metrics and VertexLabelingSolutions
    """
    
    # Start with an empty Pareto front
    pareto_front = []

    # Loop through each solution in the list
    for candidate_metric, soln in zip(all_pars_metrics, solutions):
        # Assume candidate is not dominated; check against all others
        if not any(all(other[0][i] <= candidate_metric[i] for i in range(len(candidate_metric))) and
                   any(other[0][i] < candidate_metric[i] for i in range(len(candidate_metric))) 
                   for other in pareto_front):
            # If no one in the current Pareto front dominates the candidate, add it
            pareto_front.append((candidate_metric, soln))
            # Remove any from the pareto_front that is dominated by the new candidate
            pareto_front = [front for front in pareto_front if not all(candidate_metric[i] <= front[0][i] for i in range(len(candidate_metric))) or not any(candidate_metric[i] < front[0][i] for i in range(len(candidate_metric)))]

    pareto_metrics = [front[0] for front in pareto_front]
    pareto_solutions = [front[1] for front in pareto_front]
    return pareto_metrics, pareto_solutions

def run_fitch_hartigan(v_solver, results):
    """
    Recover possible ancestral states using Fitch's algorithm and select one optimal solution

    Parameters:
    - adj_matrix: Sparse COO adjacency matrix representing the tree structure (n x n).
    - node_idx_to_observed_sites: Dictionary mapping leaf node indices to their labels.
    - root_label: The known label of the root node.

    Returns:
    - ancestral_matrix: A k x n matrix, where k is the number of labels, and n is the internal node indices.
    """
    adj_matrix = v_solver.input_T
    node_idx_to_observed_sites = v_solver.idx_to_observed_sites
    root_label = torch.argmax(v_solver.p, dim=0).item()
    
    n = adj_matrix.shape[0]
    k = v_solver.num_sites

    # Step 1: Initialize a k x n matrix to store possible labels for each node (True/False for each label)
    ancestral_matrix = torch.zeros((k, n), dtype=torch.float32)

    # Step 2: Fill in the matrix for leaf nodes and root node based on the node_idx_to_observed_sites dictionary
    for node, leaf_labels in node_idx_to_observed_sites.items():
        ancestral_matrix[:, node] = 0
        for leaf_label in leaf_labels:
            ancestral_matrix[leaf_label, node] = 1  # Set the corresponding label to True

    # For leaf nodes without any observed sites, initialize as possibly belonging to all sites
    leaf_nodes = torch.nonzero(adj_matrix.sum(dim=1) == 0).squeeze(dim=1).tolist()

    for leaf in leaf_nodes:
        if leaf not in node_idx_to_observed_sites:
            ancestral_matrix[:, leaf] = 1

    root = get_root_index(adj_matrix)
    ancestral_matrix[:, root] = 0
    ancestral_matrix[root_label, root] = 1

    # Step 3: Perform the downward pass (Fitch algorithm) for internal nodes
    def fitch_down(node):
        node = int(node)

        # Check if the node has observed sites
        has_observed_sites = node in node_idx_to_observed_sites

        # Find children of the node using the adjacency matrix
        children = torch.nonzero(adj_matrix[node]).squeeze(dim=1)

        # If the node is a leaf, return its known label set
        if len(children) == 0 and has_observed_sites:
            return ancestral_matrix[:, node]

        # Collect label sets from all children
        child_label_sets = [fitch_down(child) for child in children]

        # Perform set intersection or union based on children's labels
        intersection = torch.stack(child_label_sets).all(dim=0)
        if torch.any(intersection):  # If intersection is non-empty, use it as the set
            ancestral_matrix[:, node] = intersection
        else:  # Otherwise, use the union of the children's sets
            union = torch.stack(child_label_sets).any(dim=0)
            ancestral_matrix[:, node] = union

        # If the node has observed sites, enforce them on the computed labels
        if has_observed_sites:
            observed_sites = ancestral_matrix[:, node]
            ancestral_matrix[:, node] = ancestral_matrix[:, node] * observed_sites

        return ancestral_matrix[:, node]
    
    # Step 4: Perform Fitch's downward pass starting from the root
    fitch_down(root)
    # Reinforce root label
    ancestral_matrix[:, root] = 0
    ancestral_matrix[root_label, root] = 1
    single_soln_matrix = ancestral_matrix.clone()

    # Step 5: Perform the upward pass to choose specific states for internal nodes,
    # starting with the root label.
    def fitch_up(node, parent_label):
        children = torch.nonzero(adj_matrix[node]).squeeze(dim=1)

        # If there are multiple possible labels, choose one (use the parent's label if possible)
        possible_labels = torch.where(single_soln_matrix[:, node])[0]
        if parent_label in possible_labels:
            chosen_label = parent_label
        else:
            chosen_label = possible_labels[0]

        # Set the chosen label to True and others to False for the node
        single_soln_matrix[:, node] = 0
        single_soln_matrix[chosen_label, node] = 1

        # Pass the chosen label to the children
        for child in children:
            fitch_up(child, chosen_label)

    # Perform the upward pass starting from the root with the enforced root label
    fitch_up(root, root_label)

    # We needed to include the root labeling for Fitch-Hartigan, but we restack 
    # it using stack_vertex_labeling, so remove it momentarily
    single_soln_matrix = torch.cat((single_soln_matrix[:, :root], single_soln_matrix[:, root+1:]), dim=1)
    V = stack_vertex_labeling(v_solver.L, add_batch_dim(single_soln_matrix), v_solver.p, None, None)
    T = repeat_n(v_solver.full_T,1).to_sparse()
    metrics = ancestral_labeling_metrics(V, T, v_solver.G, v_solver.O, v_solver.p, 
                                         update_path_matrix=True, compute_full_c=True, identical_T=True)
    V = V.cpu().detach()
    T = T.cpu().detach()
    metrics = tuple(metric.cpu().detach() for metric in metrics)
    print("Fitch-hartigan result:", metrics)
    results.append((V, torch.zeros(V.shape, device=V.device), T, None, (*metrics,torch.zeros(size=(1,),device=V.device))))



from setuptools import setup, find_packages

with open('requirements.txt') as f:
    requirements = f.read().splitlines()
    #print(requirements)

setup(name='metient', version='0.1.3.4.1', url="https://github.com/divyakoyy/metient.git", packages=['metient', 'metient.util', 'metient.lib'], install_requires=requirements,)



import numpy as np
import networkx as nx
import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import torch
from networkx.drawing.nx_pydot import to_pydot
from PIL import Image as PILImage
from IPython.display import Image, display
import io
import matplotlib.gridspec as gridspec
import math
import matplotlib.font_manager
from matplotlib import rcParams
import os
import pickle
import pygraphviz as pgv
from collections import deque
import gzip
import re
import copy
import tracemalloc

import metient.util.vertex_labeling_util as vutil 
from metient.util.globals import *

import pandas as pd
pd.options.display.float_format = '{:,.3f}'.format

FONT = "Arial"

if torch.cuda.is_available():
    torch.set_default_tensor_type(torch.cuda.FloatTensor)

def is_cyclic(G):
    '''
    returns True if graph contains cycles
    '''
    num_nodes = G.size(0)
    visited = torch.zeros(num_nodes, dtype=torch.bool)
    stack = torch.zeros(num_nodes, dtype=torch.bool)

    def dfs(node):
        visited[node] = True
        stack[node] = True

        for neighbor in range(num_nodes):
            if G[node, neighbor] >= 1:
                if not visited[neighbor]:
                    if dfs(neighbor):
                        return True
                elif stack[neighbor]:
                    return True

        stack[node] = False
        return False

    for node in range(num_nodes):
        if not visited[node]:
            if dfs(node):
                return True

    return False

def site_clonality_with_G(G):
    '''
    Returns monoclonal if every site is seeded by one clone,
    else returns polyclonal.
    '''
    if torch.all(G == 0):
        return "n/a"
    return "polyclonal" if ((G > 1).any()) else  "monoclonal"

def site_clonality(V, A):
    '''
    Returns monoclonal if every site is seeded by one clone,
    else returns polyclonal.
    '''
    V, A = prep_V_A_inputs(V, A)
    G = migration_graph(V, A)
    return site_clonality_with_G(G)

def genetic_clonality(V, A, idx_to_label):
    '''
    Returns monoclonal if every site is seeded by the *same* clone,
    else returns polyclonal.
    '''
    V, A = prep_V_A_inputs(V, A)
    all_seeding_clusters = seeding_clusters(V, A, idx_to_label)
    if len(all_seeding_clusters) == 0:
        return "n/a"
    monoclonal = True if len(all_seeding_clusters) == 1 else False
    return "monoclonal" if monoclonal else "polyclonal"

def seeding_pattern_with_G(G):

    # Determine if single-source seeding (all incoming edges to a site in G 
    # originate from the same site) OR multi-source seeding (at least one site is 
    # seeded from multiple other sites) OR (R) reseeding (at least one site seeds 
    # its originating site)
    non_zero = torch.where(G > 0)
    source_sites = non_zero[0]
    binarized_G = (G != 0).to(torch.int)

    col_sums = torch.sum(binarized_G, axis=0)
    # single-source means that each site is only seeded by ONE other site 
    # (not that seeding site is 1)
    is_single_source = torch.all(col_sums <= 1).item()
    unique_source_sites = torch.unique(source_sites)

    if len(unique_source_sites) == 0:
        return "no seeding"
    elif is_cyclic(G):
        return "reseeding"
    elif len(unique_source_sites) == 1:
        return "primary single-source"
    elif is_single_source:
        return "single-source"
    return "multi-source"

def seeding_pattern(V, A):
    '''
    V: Vertex labeling matrix where columns are one-hot vectors representing the
    anatomical site that the node originated from (num_sites x num_nodes)
    A:  Adjacency matrix (directed) of the full tree (num_nodes x num_nodes)

    returns: verbal description of the seeding pattern, one of:
    {primary single-source, single-source, multi-source, reseeding}
    '''
    G = migration_graph(V, A)
    return seeding_pattern_with_G(G)
    

def migration_edges(V, A, sites=None):
    '''
    V: Vertex labeling matrix where columns are one-hot vectors representing the
    anatomical site that the node originated from (num_sites x num_nodes)
    A:  Adjacency matrix (directed) of the full tree (num_nodes x num_nodes)
    sites: optional, set of indices in range [0,num_sites) that you restrict migration edges *to*
    Returns:
        Returns a matrix where Yij = 1 if there is a migration edge from node i to node j
    '''
    X = V.T @ V 
    Y = torch.mul(A, (1-X))
    
    # Remove migration edges to sites that we're not interested in 
    if sites != None:
        Y = Y.coalesce()
        indices = Y.indices()
        values = Y.values()

        # Iterate over the sparse tensor's nonzero elements using tree_iterator
        for idx in range(indices.size(1)):
            i, j = indices[:, idx]  # Get the row (i) and column (j) indices
            k = (V[:, j] == 1).nonzero(as_tuple=True)[0][0].item()
            if k not in sites:
                values[idx] = 0  # Set the value to zero

        # Remove zeroed-out entries efficiently
        mask = values != 0  # Keep only non-zero values
        Y = torch.sparse_coo_tensor(indices[:, mask], values[mask], Y.size())
    return Y

def seeding_cluster_parents(V, A, sites=None):
    '''
    returns: list of nodes whose child is a different color
    '''
    V, A = prep_V_A_inputs(V, A)
    Y = migration_edges(V,A, sites)
    if Y.is_sparse:
        Y = Y.to_dense()
    seeding_clusters = torch.nonzero(Y.any(dim=1)).squeeze()
    # Check if it's a scalar (0D tensor)
    if seeding_clusters.dim() == 0:
        # Convert to a 1D tensor (vector)
        seeding_clusters = seeding_clusters.unsqueeze(0)
    seeding_clusters = [int(x) for x in seeding_clusters]
    return seeding_clusters

def seeding_clusters(V, A, node_idx_to_label, sites=None):
    '''
    returns: list of nodes whose parent is a different color
    '''
    V, A = prep_V_A_inputs(V, A)
    Y = migration_edges(V,A, sites)

    Y = Y.coalesce()
    nonzero_indices = Y.indices()

    # Select the indices where Y is not 0
    seeding_clusters = nonzero_indices[1][Y.values() != 0]
    # seeding_clusters = (Y == 1).nonzero(as_tuple=True)[1]
    # Check if it's a scalar (0D tensor)
    if seeding_clusters.dim() == 0:
        # Convert to a 1D tensor (vector)
        seeding_clusters = seeding_clusters.unsqueeze(0)
    seeding_clusters = [int(x) for x in seeding_clusters]

    if isinstance(node_idx_to_label, vutil.MigrationHistoryNodeCollection):
        node_collection = node_idx_to_label
    else:
        node_collection = vutil.MigrationHistoryNodeCollection.from_dict(node_idx_to_label)

    # Special case for if the node w/ diff color parent is a witness node or polytomy resolver node,
    # the seeding cluster is the parent node (since no new mutations)
    all_seeding_clusters = set()
    for x in seeding_clusters:
        node = node_collection.get_node(x)
        if node.is_witness or node.is_polytomy_resolver_node:
            all_seeding_clusters.add(vutil.get_parent(A, node.idx))
        else:
            all_seeding_clusters.add(x)
    return list(all_seeding_clusters)

def find_tree_trunk(adj_matrix):
    n = len(adj_matrix)  # Number of nodes in the matrix
    root = vutil.get_root_index(adj_matrix)  # Assuming the root node is 0
    
    # Function to find children of a given node
    def children(node):
        return [i for i in range(n) if adj_matrix[node][i] == 1]
    
    # Start from the root
    current_node = root
    trunk = [current_node]  # Initialize the trunk with the root node

    # Continue until a node has more than one child
    while True:
        children = children(current_node)
        if len(children) != 1:  # More than one child or no children
            break
        current_node = children[0]  # Move to the next node in the trunk
        trunk.append(current_node)
    
    return trunk

def is_valid_path(path, S):
    return all(node in path for node in S)

def mrca(adj_matrix, nodes_to_check):
    '''
    Gets the most recent common ancestor of nodes in nodes_to_check
    '''
    start_node = vutil.get_root_index(adj_matrix)
    num_nodes = len(adj_matrix)
    visited = [False] * num_nodes

    queue = deque()
    queue.append(start_node)
    visited[start_node] = True

    while queue:
        current_node = queue.popleft()
        if current_node in nodes_to_check:
            return current_node

        for neighbor, connected in enumerate(adj_matrix[current_node]):
            if connected and not visited[neighbor]:
                queue.append(neighbor)
                visited[neighbor] = True
                
def hamiltonian_paths(adj_matrix, path, visited, n, S):
    if is_valid_path(path, S):
        return True

    current_node = path[-1]
    for next_node in range(n):
        if adj_matrix[current_node][next_node] == 1 and not visited[next_node]:
            visited[next_node] = True
            path.append(next_node)
            
            if hamiltonian_paths(adj_matrix, path, visited, n, S):
                return True

            path.pop()
            visited[next_node] = False
            
    return False

def has_hamiltonian_path_with_set(adj_matrix, nodes_to_check):
    n = len(adj_matrix)
    highest_node = mrca(adj_matrix, nodes_to_check)
    visited = [False] * n
    path = [highest_node]
    visited[highest_node] = True

    if hamiltonian_paths(adj_matrix, path, visited, n, nodes_to_check):
        return True
    
    return False
                 
def phyleticity(V, A, idx_to_label, sites=None):
    '''
    If all nodes can be reached from the top level node in the seeding clusters,
    returns monophyletic, else polyphyletic
    '''
    
    V, A = prep_V_A_inputs(V, A)
    clonality = genetic_clonality(V, A, idx_to_label)
    all_seeding_clusters = seeding_clusters(V, A, idx_to_label, sites)
    
    if len(all_seeding_clusters) == 0: # no seeding
        return "n/a"
    
    if "monoclonal" in clonality:
        return "monophyletic"

    num_nodes = len(A)
    visited = [False] * num_nodes
    highest_node = mrca(A, all_seeding_clusters)
    
    # Do a single DFS from highest_node to build reachable set
    num_nodes = len(A)
    visited = [False] * num_nodes
    reachable = set()
    
    # Stack-based DFS is faster than recursive
    stack = [highest_node]
    while stack:
        node = stack.pop()
        if not visited[node]:
            visited[node] = True
            reachable.add(node)
            # Add unvisited neighbors to stack
            for neighbor, connected in enumerate(A[node]):
                if connected and not visited[neighbor]:
                    stack.append(neighbor)
    
    # Check if all seeding clusters are reachable
    if all(node in reachable for node in all_seeding_clusters):
        return "monophyletic"
    return "polyphyletic"

def tracerx_phyleticity(V, A, idx_to_label):
    '''
    Looking at the seeding clones, is there a single path that connects all seeding clusters
    or multiple possible paths? If singular path, monophyletic, if not, polyphyletic

    This is to implement TRACERx's definition of phyleticity:
    "the origin of the seeding clusters was determined as monophyletic if all 
    clusters appear along a single branch, and polyphyletic if clusters were
    spread across multiple branches of the phylogenetic tree. Thus, if a 
    metastasis was defined as monoclonal, the origin was necessarily monophyletic. 
    For polyclonal metastases, the clusters were mapped to branches of the 
    evolutionary tree. If multiple branches were found, the origin was determined 
    to be polyphyletic, whereas, if only a single branch gave rise to all shared 
    clusters, the origin was defined as monophyletic."
    (https://www.nature.com/articles/s41586-023-05729-x#Sec7)
    '''
    V, A = prep_V_A_inputs(V, A)
    clonality = genetic_clonality(V, A, idx_to_label)
    all_seeding_clusters = seeding_clusters(V, A, idx_to_label)

    is_hamiltonian = has_hamiltonian_path_with_set(A, all_seeding_clusters)
    if "monoclonal" in clonality:
        return "monophyletic"
    phyleticity = "monophyletic" if is_hamiltonian else "polyphyletic"
    return phyleticity
    
def tracerx_seeding_pattern(V, A, idx_to_label):
    '''
    V: Vertex labeling matrix where columns are one-hot vectors representing the
    anatomical site that the node originated from (num_sites x num_nodes)
    A:  Adjacency matrix (directed) of the full tree (num_nodes x num_nodes)

    Monoclonal if only one clone seeds met(s), else polyclonal
    Monophyletic if there is a Hamiltonian path connecting all seeding clusters
    returns: one of {no seeding, monoclonal monophyletic, polyclonal polyphyletic, polyclonal monophyletic}
    '''
    V, A = prep_V_A_inputs(V, A)
    G = migration_graph(V, A)
    non_zero = torch.where(G > 0)
    source_sites = non_zero[0]
    if len(torch.unique(source_sites)) == 0:
        return "no seeding"

    # 1) determine if monoclonal (only one clone seeds met(s)), else polyclonal
    clonality = genetic_clonality(V,A,idx_to_label)
    
    # 2) determine if monophyletic or polyphyletic
    phyleticity = tracerx_phyleticity(V, A,idx_to_label)

    return clonality + phyleticity

def write_tree(tree_edge_list, output_filename, add_germline_node=False):
    '''
    Writes the full tree to file like so:
    0 1
    1 2;3
    '''
    if add_germline_node:
        tree_edge_list.append(('GL', tree_edge_list[0][0]))
    with open(output_filename, 'w') as f:
        for edge in tree_edge_list:
            f.write(f"{edge[0]} {edge[1]}")
            f.write("\n")

def write_tree_vertex_labeling(vertex_name_to_site_map, output_filename, add_germline_node=False):
    '''
    Writes the full tree's vertex labeling to file like so:
    1 P
    1_P P
    25;32_M1 M1
    '''
    if add_germline_node:
        vertex_name_to_site_map['GL'] = "P"
    with open(output_filename, 'w') as f:
        for vert_label in vertex_name_to_site_map:
            f.write(f"{vert_label} {vertex_name_to_site_map[vert_label]}")
            f.write("\n")

def write_migration_graph(migration_edge_list, output_filename):
    '''
    Writes the full migration graph to file like so:
    P M1
    P M2
    P M1
    M1 M2
    '''
    with open(output_filename, 'w') as f:
        for edge in migration_edge_list:
            f.write(f"{edge[0]} {edge[1]}")
            f.write("\n")

def plot_losses(losses):
    plt.plot([x for x in range(len(losses))],losses, label="loss")
    plt.xlabel("epoch")
    plt.ylabel("loss")
    plt.show()

def plot_temps(temps):
    plt.plot([x for x in range(len(temps))],temps, label="temp")
    plt.xlabel("epoch")
    plt.ylabel("temp")
    plt.show()

def plot_loss_components(loss_dicts, weights):
    # if else statements to handle lr schedules where we do not calculate
    # all loss components at every epoch

    mig_losses = [e[MIG_KEY] for e in loss_dicts]
    seed_losses = [e[SEEDING_KEY] for e in loss_dicts]
    neg_entropy = [e[ENTROPY_KEY] for e in loss_dicts]

    plt.plot([x for x in range(len(loss_dicts))],mig_losses, label="m")
    plt.plot([x for x in range(len(loss_dicts))],seed_losses, label="s")
    plt.plot([x for x in range(len(loss_dicts))],neg_entropy, label="neg. ent.")

    plt.xlabel("epoch")
    plt.ylabel("loss")
    plt.legend(loc="upper right")
    plt.show()

def view_pydot(pdot):
    plt = Image(pdot.create_png())
    display(plt)

def contains_delim(s, delims):
    for delim in delims:
        if delim in s:
            return True
    return False

def pruned_mut_label(mut_names, shorten_label, to_string):
    if not shorten_label and not to_string:
        return ([str(m) for m in mut_names])
    elif not shorten_label and to_string:
        return ";".join([str(m) for m in mut_names])
    # If mutation name contains :, ;, _ (e.g. LOC1:9:123), take everything before the first colon for display
    delims = [":", ";", "_"]
    gene_names = []
    for mut_name in mut_names:
        mut_name = str(mut_name)
        if not contains_delim(mut_name, delims):
            gene_names.append(mut_name)
        else:
            gene_names.append(re.split(r"[_;:]", mut_name)[0])
    # Try to find relevant cancer genes to label
    gene_candidates = set()
    for gene in gene_names:
        gene = gene.upper()
        if gene in CANCER_DRIVER_GENES:
            gene_candidates.add(gene)
        elif gene in ENSEMBLE_TO_GENE_MAP:
            gene_candidates.add(ENSEMBLE_TO_GENE_MAP[gene])
    final_genes = gene_names if len(gene_candidates) == 0 else gene_candidates
   
    k = 2 if len(final_genes) > 2 else len(final_genes)
    if to_string:
        return ";".join(list(final_genes)[:k])
    else:
        return list(final_genes)

def full_tree_node_idx_to_label(T, node_collection, shorten_label=True, to_string=False):
    '''
    Build a map of node_idx to (label, is_witness, is_polytomy_resovler_node) for plotting and saving
    information to pickle file
    '''

    full_node_idx_to_label_map = dict()
    for i, j in vutil.tree_iterator(T):
        node_i = node_collection.get_node(i)
        label = pruned_mut_label(node_i.label, shorten_label, to_string)
        full_node_idx_to_label_map[i] = (label, node_i.is_witness, node_i.is_polytomy_resolver_node)
        
        node_j = node_collection.get_node(j)
        label = pruned_mut_label(node_j.label, shorten_label, to_string)
        full_node_idx_to_label_map[j] = (label, node_j.is_witness, node_j.is_polytomy_resolver_node)
    return full_node_idx_to_label_map

def idx_to_color(custom_colors, idx, alpha=1.0):
    rgb = mcolors.to_rgb(custom_colors[idx])
    rgb_alpha = (rgb[0], rgb[1], rgb[2], alpha)
    return mcolors.to_hex(rgb_alpha, keep_alpha=True)

def prep_V_A_inputs(V, A):
    if not isinstance(V, torch.Tensor):
        V = torch.tensor(V)
    if not isinstance(A, torch.Tensor):
        A = torch.tensor(A)
    return V, A

def migration_graph(V, A):
    '''
    V: Vertex labeling matrix where columns are one-hot vectors representing the
    anatomical site that the node originated from (num_sites x num_nodes)
    A:  Adjacency matrix (directed) of the full tree (num_nodes x num_nodes)
    '''
    V, A = prep_V_A_inputs(V, A)
    migration_graph = (V @ A) @ V.T
    migration_graph_no_diag = torch.mul(migration_graph, 1-torch.eye(migration_graph.shape[0], migration_graph.shape[1], device=migration_graph.device))
    
    return migration_graph_no_diag

def find_abbreviation_mark(input_string):
    abbreviation_marks = [',', '-', '|']  # Add other abbreviation marks as needed

    for mark in abbreviation_marks:
        if input_string.count(mark) == 1:
            return mark
    return None

def migration_graph_dot(V, A, ordered_sites, custom_colors, show=True):
    '''
    Plots migration graph G which represents the migrations/comigrations between
    all anatomical sites.

    Returns a list of edges (e.g. [('P' ,'M1'), ('P', 'M2')])
    '''
    # Reformat anatomical site strings if too long for display and there
    # is an easy way to split the string (abbreviation of some sort)
    fmted_ordered_sites = []
    for site in ordered_sites:
        if len(site) > 17:
            mark = find_abbreviation_mark(site)
            if mark != None:
                fmted_ordered_sites.append(f"{mark}\n".join(site.split(mark)))
            else:
                fmted_ordered_sites.append(site)
        else:
            fmted_ordered_sites.append(site)

    mig_graph_no_diag = migration_graph(V, A)

    G = nx.MultiDiGraph()
    for node, color in zip(fmted_ordered_sites, custom_colors):
        G.add_node(node, shape="box", color=color, fillcolor='white', fontname=FONT, penwidth=3.0)

    for i, adj_row in enumerate(mig_graph_no_diag):
        for j, num_edges in enumerate(adj_row):
            if num_edges > 0:
                for _ in range(int(num_edges.item())):
                    G.add_edge(fmted_ordered_sites[i], fmted_ordered_sites[j], color=f'"{custom_colors[i]};0.5:{custom_colors[j]}"', penwidth=3)

    dot = nx.nx_pydot.to_pydot(G)
    if show:
        view_pydot(dot)

    dot_lines = dot.to_string().split("\n")
    dot_lines.insert(1, 'dpi=600;size=3.5;')
    dot_str = ("\n").join(dot_lines)

    return dot_str

def migration_history_tree_dot(V, T, gen_dist, custom_colors, node_collection=None, show=True):

    # (1) Create full directed graph 
    # these labels are used for display in plotting
    display_node_idx_to_label_map = full_tree_node_idx_to_label(T, node_collection, shorten_label=True, to_string=True)            

    color_map = { i:idx_to_color(custom_colors, (V[:,i] == 1).nonzero()[0][0].item()) for i in range(V.shape[1])}
    G = nx.DiGraph()
    node_options = {"label":"", "shape": "circle", "penwidth":3, 
                    "fontname":FONT, "fontsize":12,
                    "fixedsize":"true", "height":0.25}

    if gen_dist != None:
        gen_dist = ((gen_dist / torch.max(gen_dist[gen_dist>0]))*1.5) + 1

    for i, j in vutil.tree_iterator(T):
        label_i, _, _ = display_node_idx_to_label_map[i]
        label_j, is_witness, _ = display_node_idx_to_label_map[j]
        
        # label_i, label_j = "",""
        G.add_node(i, xlabel=label_i, fillcolor=color_map[i], 
                    color=color_map[i], style="filled", **node_options)
        G.add_node(j, xlabel="" if is_witness else label_j, fillcolor=color_map[j], fontcolor="white" if is_witness else "black",
                    color=color_map[j], style="solid" if is_witness else "filled", **node_options)

        style = "dashed" if is_witness else "solid"
        penwidth = 5 if is_witness else 5.5

        # We're iterating through i,j of the full tree (including leaf nodes),
        # while G only has genetic distances between internal nodes
        minlen = gen_dist[i, j].item() if (gen_dist != None and i < len(gen_dist) and j < len(gen_dist)) else 1.0

        G.add_edge(i, j,color=f'"{color_map[i]};0.5:{color_map[j]}"', 
                   penwidth=penwidth, arrowsize=0, style=style, minlen=minlen)

    # Add edge from normal to root 
    root_idx = vutil.get_root_index(T)
    root_label = display_node_idx_to_label_map[root_idx][0]
    G.add_node("normal", label="", xlabel=root_label, penwidth=3, style="invis")
    G.add_edge("normal", root_idx, label="", 
                color=f'"{color_map[root_idx]}"', 
                penwidth=4, arrowsize=0, style="solid")

    assert(nx.is_tree(G))

    # we have to use graphviz in order to get multi-color edges :/
    dot = to_pydot(G)
    dot.set_graph_defaults(layout='dot', seed=42) 
    dot = dot.to_string().split("\n")
    # hack since there doesn't seem to be API to modify graph attributes...
    # dot.insert(1, 'graph[splines=false]; nodesep=0.7; rankdir=TB; ranksep=0.6; forcelabels=true; dpi=800; size=2.5;')
    dot.insert(1, 'graph[splines=false]; nodesep=0.4; rankdir=TB; ranksep=0.4; forcelabels=true; dpi=800; size=2.5; seed=42')
    dot_str = ("\n").join(dot)

    if show:
        dot = nx.nx_pydot.to_pydot(dot_str)
        view_pydot(dot)

    return dot_str

def collect_top_tree_info(V, T, node_collection, ordered_sites):
    '''
    Get info about the best migration history to return to the user
    '''
    # these labels are used for writing out full vertex names to file
    full_node_idx_to_label_map = full_tree_node_idx_to_label(T, node_collection, shorten_label=False, to_string=False)
    tree_edges = []
    for i, j in vutil.tree_iterator(T):
        tree_edges.append((full_node_idx_to_label_map[i][0], full_node_idx_to_label_map[j][0]))
    vertex_name_to_site_map = { ";".join(full_node_idx_to_label_map[i][0]):ordered_sites[(V[:,i] == 1).nonzero()[0][0].item()] for i in range(V.shape[1])}

    mig_graph_no_diag = migration_graph(V, T)

    mig_edges = []
    for i, adj_row in enumerate(mig_graph_no_diag):
        for j, num_edges in enumerate(adj_row):
            if num_edges > 0:
                for _ in range(int(num_edges.item())):
                    mig_edges.append((ordered_sites[i], ordered_sites[j]))

    return tree_edges, full_node_idx_to_label_map, vertex_name_to_site_map, mig_edges

def construct_loss_dict(soln, full_loss):
    loss_dict = {MIG_KEY: soln.m, COMIG_KEY:soln.c, SEEDING_KEY: soln.s, ORGANOTROP_KEY: soln.o, GEN_DIST_KEY: soln.g, ENTROPY_KEY: soln.e}
    loss_dict = {**loss_dict, **{FULL_LOSS_KEY: round(torch.mean(full_loss).item(), 3)}}
    return loss_dict

def convert_lists_to_np_arrays(pickle_outputs, keys):
    '''
    makes unpickling these outputs much faster
    '''
    for key in keys:
        if key in pickle_outputs:
            pickle_outputs[key] = np.array(pickle_outputs[key])

    return pickle_outputs

def figure_output_pattern(V, A, idx_to_label):
    '''
    Display clonality and phyleticity
    '''
    gen_clonality = genetic_clonality(V, A, idx_to_label).replace("clonal", "")
    st_clonality = site_clonality(V, A).replace("clonal", "")
    pattern = seeding_pattern(V, A)
    phyletic = phyleticity(V, A, idx_to_label)
    output_str = f"{pattern}, {phyletic}\n"
    output_str += f"genetic clonality: {gen_clonality}, site clonality: {st_clonality}\n"
    return output_str

def get_components(tensor):
    """Get components to pickle a sparse COO tensor."""
    if not tensor.is_sparse:
        return tensor.detach().cpu().numpy()

    # Extract components
    indices = tensor._indices()
    values = tensor._values()
    size = tensor.size()

    # Pack components into a tuple
    return (indices, values, size)

def dense_to_sparse(dense_tensor):
    """Convert a dense tensor to a sparse COO tensor and delete the dense tensor."""
    # Ensure the input is a 2D tensor (you can modify for N-D tensors if needed)
    if dense_tensor.dim() != 2:
        raise ValueError("Input tensor must be 2-dimensional.")
    
    # Get the indices of non-zero elements
    indices = torch.nonzero(dense_tensor, as_tuple=False).t()
    
    # Get the values of the non-zero elements
    values = dense_tensor[indices[0], indices[1]]

    # Create a sparse COO tensor
    sparse_tensor = torch.sparse_coo_tensor(indices, values, size=dense_tensor.size())
    
    # Delete the dense tensor to free memory
    del dense_tensor

    return sparse_tensor

def save_best_trees(min_loss_solutions, U, O, weights, ordered_sites, print_config, primary, output_dir, run_name, original_root_idx=-1):
    '''
    min_loss_solutions is in order from lowest to highest loss 

    original_root_idx: if not -1, swap the original_root_idx with 0 in all
    data that we save that involves node/cluster indices. This will then match
    the inputs from the user's again.
    '''
    
    primary_idx = ordered_sites.index(primary)

    ret = None
    figure_outputs = []
    pickle_outputs = {OUT_LABElING_KEY:[], OUT_LOSSES_KEY:[],OUT_IDX_LABEL_KEY:[],
                      OUT_ADJ_KEY:[], OUT_SITES_KEY:ordered_sites, OUT_LOSS_DICT_KEY:[],
                      OUT_PRIMARY_KEY:primary, 
                      OUT_SOFTV_KEY:[], OUT_GEN_DIST_KEY:[]}

    with torch.no_grad():
        if print_config.custom_colors == None:
            custom_colors = copy.deepcopy(DEFAULT_COLORS)
            # Reorder so that green is always the primary
            green_idx = custom_colors.index(DEFAULT_GREEN)
            custom_colors[primary_idx], custom_colors[green_idx] = custom_colors[green_idx], custom_colors[primary_idx]
        else:
            custom_colors = print_config.custom_colors

        for i, min_loss_solution in enumerate(min_loss_solutions):
            V = min_loss_solution.V
            soft_V = min_loss_solution.soft_V
            T = min_loss_solution.T
            G = min_loss_solution.G
            full_loss = min_loss_solution.loss
            node_collection = min_loss_solution.node_collection
            loss_dict = construct_loss_dict(min_loss_solution, full_loss)
            
            # Restructure adjacency matrices so that node indices match the cluster indices
            # that the user originally input (we restructure them s.t. root index is 0 during
            # inference to make indexing logic much simpler)
            # TODO: this is unideal, restructure matrices should take in sparse matrix
            # from here on, return a parents matrix?
            needs_sparse_conversion = False
            if original_root_idx != -1:
                if T.is_sparse:
                    T = T.to_dense()
                    needs_sparse_conversion = True
                node_collection = copy.deepcopy(node_collection)
                T, _, _, node_collection, G, _, V, U = vutil.restructure_matrices(0, original_root_idx, T, None, None, node_collection, G, None, V, U)
            if needs_sparse_conversion:
                T = dense_to_sparse(T)
                
            edges, full_tree_idx_to_label, vertices_to_sites_map, mig_graph_edges = collect_top_tree_info(V, T, node_collection, ordered_sites)
                
            tree_dot = migration_history_tree_dot(V, T, G, custom_colors, node_collection, show=False)
            mig_graph_dot = migration_graph_dot(V, T, ordered_sites, custom_colors, show=False)

            if print_config.visualize:
                pattern = figure_output_pattern(V, T, full_tree_idx_to_label)
            else:
                pattern = ""

            figure_outputs.append((tree_dot, mig_graph_dot, loss_dict, pattern))
            pickle_outputs[OUT_LABElING_KEY].append(V.detach().cpu().numpy())
            pickle_outputs[OUT_LOSSES_KEY].append(full_loss.cpu().numpy())
            
            pickle_outputs[OUT_ADJ_KEY].append(get_components(T))
            pickle_outputs[OUT_SOFTV_KEY].append(soft_V.detach().cpu().numpy())
            pickle_outputs[OUT_OBSERVED_CLONES_KEY] = U.detach().cpu().numpy() if U != None else np.array([])
        
            if G != None:
                pickle_outputs[OUT_GEN_DIST_KEY].append(G.detach().cpu().numpy())                
            pickle_outputs[OUT_LOSS_DICT_KEY].append(loss_dict)
            pickle_outputs[OUT_IDX_LABEL_KEY].append(full_tree_idx_to_label)
            if i == 0: # Best tree
                ret = (edges, vertices_to_sites_map, mig_graph_edges, loss_dict)

        #pickle_outputs = convert_lists_to_np_arrays(pickle_outputs, [OUT_LABElING_KEY, OUT_LOSSES_KEY, OUT_ADJ_KEY, OUT_SOFTV_KEY, OUT_GEN_DIST_KEY])

        save_outputs(figure_outputs, print_config, output_dir, run_name, pickle_outputs, weights)

    return ret

def formatted_loss_string(loss_dict, weights):
    s = f"Loss: {loss_dict[FULL_LOSS_KEY]}\n\n"

    s += f"Migration num.: {int(loss_dict[MIG_KEY])}\n"
    s += f"Comigration num.: {int(loss_dict[COMIG_KEY])}\n"
    s += f"Seeding site num.: {int(loss_dict[SEEDING_KEY])}\n"
    s += f"Neg. entropy: {round(float(loss_dict[ENTROPY_KEY]), 3)}\n"

    if weights.gen_dist != 0:
        s += f"Genetic dist. loss: {round(float(loss_dict[GEN_DIST_KEY]), 3)}\n"
    if weights.organotrop != 0:
        s += f"Organotrop. loss: {round(float(loss_dict[ORGANOTROP_KEY]), 3)}\n"
    return s

def save_outputs(figure_outputs, print_config, output_dir, run_name, pickle_outputs, weights):

    if print_config.visualize:
        k = print_config.k_best_trees
        sys_fonts = matplotlib.font_manager.findSystemFonts(fontpaths=None, fontext='ttf')
        for font in sys_fonts:
            if FONT in font:
                matplotlib.font_manager.fontManager.addfont(font)
                rcParams['font.family'] = FONT

        n = len(figure_outputs)
        print(run_name)
        tense = "were" if n > 1 else "was"
        print(f"{n} unique solutions {tense} found.")

        max_trees = 20
        if n > max_trees:
            print(f"More than {max_trees} solutions detected, only plotting top {max_trees} trees.")
            n = max_trees
        # Create a figure and subplots
        #fig, axs = plt.subplots(3, k*2, figsize=(10, 8))

        plt.suptitle(run_name)

        z = 2 # number of trees displayed per row

        nrows = math.ceil(n/z)
        h = nrows*4
        fig = plt.figure(figsize=(8,h))
        
        vspace = 1/nrows

        for i, (tree_dot, mig_graph_dot, loss_info, seeding_pattern) in enumerate(figure_outputs):
            if i >= max_trees:
                break
            tree = pgv.AGraph(string=tree_dot).draw(format="png", prog="dot", args="-Glabel=\"\"")
            tree = PILImage.open(io.BytesIO(tree))
            mig_graph = pgv.AGraph(string=mig_graph_dot).draw(format="png", prog="dot")
            mig_graph = PILImage.open(io.BytesIO(mig_graph))

            gs = gridspec.GridSpec(3, 1, height_ratios=[0.02, 0.73, 0.25])

            row = math.floor(i/2)
            pad = 0.02

            # left = 0.0 if i is odd, 0.55 if even
            # right = 0.45 if i is odd, 1.0 if even
            gs.update(left=0.0+((i%2)*0.53), right=0.47+0.55*(i%2), top=1-(row*vspace)-pad, bottom=1-((row+1)*vspace)+pad, wspace=0.05)

            # Top row: Title
            ax1 = plt.subplot(gs[0])
            ax1.text(0.5, 0.5, f'Solution {i+1}\n{seeding_pattern}', ha='center', va='center', fontsize=7)
            ax1.axis('off')  # Hide the axis

            # Second row: Plot for the tree
            ax2 = plt.subplot(gs[1])
            ax2.imshow(tree)
            ax2.axis('off')

            # Third row: Create a subgrid for the migration graph and loss information
            gs_bottom = gridspec.GridSpecFromSubplotSpec(1, 2, subplot_spec=gs[2], wspace=0.05)

            # Left column for the migration graph
            ax3 = plt.subplot(gs_bottom[0])
            ax3.imshow(mig_graph)
            ax3.axis('off')

            # Right column for loss information
            ax4 = plt.subplot(gs_bottom[1])
            ax4.text(0.5, 0.5, formatted_loss_string(loss_info, weights), ha='center', va='center', fontsize=7)
            ax4.axis('off')

        fig1 = plt.gcf()
        plt.show()
        plt.close()
        if print_config.save_outputs: 
            fig1.savefig(os.path.join(output_dir, f'{run_name}.png'), dpi=600, bbox_inches='tight')

    if print_config.save_outputs:
        if not os.path.isdir(output_dir):
            raise ValueError(f"{output_dir} does not exist.")
        if print_config.verbose: print(f"Saving {run_name} to {output_dir}")
        # Save results to pickle file
        # with open(os.path.join(output_dir, f"{run_name}.pickle"), 'wb') as handle:
        with gzip.open(os.path.join(output_dir,f"{run_name}.pkl.gz"), 'wb') as gzip_file:
            pickle.dump(pickle_outputs, gzip_file, protocol=pickle.HIGHEST_PROTOCOL)
            
        # Save best dot to file
        tree_dot, mig_graph_dot, _, _ = figure_outputs[0]
        with open(os.path.join(output_dir, f"{run_name}.tree.dot"), 'w') as file:
            file.write(tree_dot)
        with open(os.path.join(output_dir, f"{run_name}.mig_graph.dot"), 'w') as file:
            file.write(mig_graph_dot)


import torch
from metient.util import vertex_labeling_util as vutil
from torch.distributions.binomial import Binomial
import numpy as np

class ObservedClonesSolver:
    def __init__(self, num_sites, num_internal_nodes, ref, var, omega, idx_to_observed_sites,
                 input_T, G, node_collection, weights, config, estimate_observed_clones, ordered_sites):
        self.ref = ref
        self.var = var
        self.omega = omega
        self.input_T = input_T
        self.G = G
        self.weights = weights
        self.config = config
        self.num_sites = num_sites
        self.num_internal_nodes = num_internal_nodes
        self.node_collection = node_collection
        self.estimate_observed_clones = estimate_observed_clones
        self.idx_to_observed_sites = idx_to_observed_sites
        self.ordered_sites = ordered_sites
    
    def run(self):
        if not self.estimate_observed_clones:
            T, G, L, node_collection = vutil.full_adj_matrix_from_internal_node_idx_to_sites_present(self.input_T, self.G, self.idx_to_observed_sites, 
                                                                                                     self.num_sites, self.config['identical_clone_gen_dist'],
                                                                                                     self.node_collection, self.ordered_sites)
            return None, self.input_T, T, G, L, node_collection, self.num_internal_nodes, self.idx_to_observed_sites

        return find_umap(self)

def find_umap(u_solver):
    
    # We're learning eta, which is the mixture matrix U (U = softmax(eta)), and tells us the existence
    # and anatomical locations of the extant clones (U > U_CUTOFF)
    #eta = -1 * torch.rand(num_sites, num_internal_nodes + 1) # an extra column for normal cells
    eta = torch.ones(u_solver.num_sites, u_solver.num_internal_nodes + 1) # an extra column for normal cells
    eta.requires_grad = True 
    u_optimizer = torch.optim.Adam([eta], lr=u_solver.config['lr'])

    B = vutil.mutation_matrix_with_normal_cells(u_solver.input_T)
    i = 0
    u_prev = eta
    u_diff = 1e9
    # losses = []
    # nlls, regs = [], []
    while u_diff > 1e-6 and i < 300:
        u_optimizer.zero_grad()
        U, u_loss, nll, reg = compute_u_loss(eta, u_solver.ref, u_solver.var, u_solver.omega, B, u_solver.weights)
        u_loss.backward()
        u_optimizer.step()
        u_diff = torch.abs(torch.norm(u_prev - U))
        u_prev = U
        i += 1
        # losses.append(u_loss.detach().numpy())
        # nlls.append(nll.detach().numpy())
        # regs.append(reg.detach().numpy())

    # plt.plot([i for i in range(len(nlls))], nlls, marker='o'); plt.show(); plt.close()
    # plt.plot([i for i in range(len(regs))], regs, marker='o'); plt.show(); plt.close()

    with torch.no_grad():
        full_T, full_G, L, idx_to_observed_sites, node_collection = vutil.full_adj_matrix_using_inferred_observed_clones(U, u_solver.input_T, u_solver.G, u_solver.num_sites, 
                                                                                                                         u_solver.config['identical_clone_gen_dist'],
                                                                                                                         u_solver.node_collection, u_solver.ordered_sites)
        # Remove any leaf nodes that aren't detected at > U_CUTOFF in any sites. These are not well estimated
        U_clones = U[:,1:]
        removal_indices = []
        for node_idx in range(U_clones.shape[1]):
            children = vutil.get_child_indices(u_solver.input_T, [node_idx])
            if node_idx not in idx_to_observed_sites and len(children) == 0:
                removal_indices.append(node_idx)
        print("node indices not well estimated", removal_indices)

        U, input_T, T, G, node_collection, idx_to_observed_sites = vutil.remove_leaf_indices_not_observed_sites(removal_indices, U, u_solver.input_T, 
                                                                                                                full_T, full_G, node_collection, idx_to_observed_sites)
        num_internal_nodes = u_solver.num_internal_nodes - len(removal_indices)
        
    return U, input_T, T, G, L, node_collection, num_internal_nodes, idx_to_observed_sites

# Adapted from PairTree
def calc_llh(F_hat, R, V, omega_v):
    '''
    Args:
        F_hat: estimated subclonal frequency matrix (num_nodes x num_mutation_clusters)
        R: Reference allele count matrix (num_samples x num_mutation_clusters)
        V: Variant allele count matrix (num_samples x num_mutation_clusters)
    Returns:
        Data fit using the Binomial likelihood (p(x|F_hat)). See PairTree (Wintersinger et. al.)
        supplement section 2.2 for details.
    '''

    N = R + V
    S, K = F_hat.shape

    for matrix in V, N, omega_v:
        assert(matrix.shape == (S, K-1))

    P = torch.mul(omega_v, F_hat[:,1:])

    bin_dist = Binomial(N, P)
    F_llh = bin_dist.log_prob(V) / np.log(2)
    assert(not torch.any(F_llh.isnan()))
    assert(not torch.any(F_llh.isinf()))

    llh_per_sample = -torch.sum(F_llh, axis=1) / S
    nlglh = torch.sum(llh_per_sample) / (K-1)
    return nlglh

def compute_u_loss(eta, ref, var, omega, B, weights):
    '''
    Args:
        eta: raw values we are estimating of matrix U (num_sites x num_internal_nodes)
        ref: Reference matrix (num_anatomical_sites x num_mutation_clusters). Num. reads that map to reference allele
        var: Variant matrix (num_anatomical_sites x num_mutation_clusters). Num. reads that map to variant allele
        omega: VAF to subclonal frequency correction 
        B: Mutation matrix (shape: num_internal_nodes x num_mutation_clusters)
        weights: Weights object

    Returns:
        Loss to score the estimated proportions of each clone in each site
    '''
    
    # Using the softmax enforces that the row sums are 1, since the proprtions of
    # clones in a given site should sum to 1
    U = torch.softmax(eta, dim=1)
    # print("eta", eta)
    #print("U", U)

    # 1. Data fit
    F_hat = (U @ B)
    nlglh = calc_llh(F_hat, ref, var, omega)
    # 2. Regularization to make some values of U -> 0
    reg = torch.sum(eta) # l1 norm 
    clone_proportion_loss = (weights.data_fit*nlglh + weights.reg*reg)
    
    return U, clone_proportion_loss, weights.data_fit*nlglh, weights.reg*reg